"""
Implements Stacked Alternating Offers (SAO) mechanism.
"""
import itertools
import math
import random
import sys
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from .common import SAOResponse, SAOState
from ..common import AgentMechanismInterface, MechanismState
from ..events import Event
from ..mechanisms import Mechanism, MechanismRoundResult
from ..outcomes import Outcome, ResponseType, outcome_is_complete
from ..utilities import UtilityFunction

__all__ = [
    "SAOMechanism",
    "SAOProtocol",
]


class SAOMechanism(Mechanism):
    """
    Implements Several variants of the Stacked Alternating Offers Protocol

    Args:
        issues: A list of issues defining the outcome-space of the negotiation (See `outcomes`).
                Only one of `issues` and `outcomes` can be passed (the other must be `None`)
        outcomes: A list of outcomes defining the outcome-space of the negotiation (See `issues`).
                  Only one of `issues` and `outcomes` can be passed (the other must be `None`)
        n_steps: The maximum number of negotiaion rounds (see `time_limit` )
        time_limit: The maximum wall-time allowed for the negotiation (see `n_steps`)
        step_time_limit: The maximum wall-time allowed for a single negotiation round.
        max_n_agents: The maximum number of negotiators allowed to join the negotiation.
        dynamic_entry: Whether it is allowed for negotiators to join the negotiation after it starts
        keep_issue_names: DEPRICATED (use `outcome_type` instead). If true, `outcomes` and `discrete_outcomes` will be
                          dictionaries otherwise they will be tuples
        outcome_type: The outcome type to use for `outcomes` and `distrete_outcomes`. It can be `tuple`, `dict` or any
                      `OutcomeType`
        cache_outcomes: If true, the mechnism will catch `outcomes` and a discrete version (`discrete_outcomes`) that
                        can be accessed by any negotiator through their AMI.
        max_n_outcomes: Maximum number or outcomes to use when disctetizing the outcome-space
        annotation: A key-value mapping to keep around. Accessible through the AMI but not used by the mechanism.
        end_on_no_response: End the negotiation if any negotiator returns NO_RESPONSE from `respond`/`counter` or returns
                            REJECT_OFFER then refuses to give an offer (by returning `None` from `proposee/`counter`).
        publish_proposer: Put the last proposer in the state.
        publish_n_acceptances: Put the number of acceptances an offer got so far in the state.
        enable_callbacks: Enable callbacks like on_round_start, etc. Note that on_negotiation_end is always received
                          by the negotiators no matter what is the setting for this parameter.
        avoid_ultimatum: If true, a proposal is taken from every agent in the first round then all of them are discarded
                         except one (choose randomly) and the algorithm continues using this one. This prevents negotiators
                         from knowing their order in the round.
        check_offers: If true, offers are checked to see if they are valid for the negotiation
                      outcome-space and if not the offer is considered None which is the same as
                      refusing to offer (NO_RESPONSE).
        ignore_negotiator_exceptions: just silently ignore negotiator exceptions and consider them no-responses.
        offering_is_accepting: Offering an outcome implies accepting it. If not, the agent who proposed an offer will
                               be asked to respond to it after all other agents.
        name: Name of the mecnanism
        **kwargs: Extra paramters passed directly to the `Mechanism` constructor

    Remarks:

        - If both `n_steps` and `time_limit` are passed, the negotiation ends when either of the limits is reached.
        - Negotiations may take longer than `time_limit` because negotiators are not interrupted while they are
          executing their `respond` or `propose` methods.

    Events:

        - negotiator_exception: Data=(negotiator, exception) raised whenever a negotiator raises an exception if
          ignore_negotiator_exceptions is set to True.
    """

    def __init__(
        self,
        issues=None,
        outcomes=None,
        n_steps=None,
        time_limit=None,
        step_time_limit=None,
        max_n_agents=None,
        dynamic_entry=True,
        keep_issue_names=None,
        outcome_type=tuple,
        cache_outcomes=True,
        max_n_outcomes: int = 1000000,
        annotation: Optional[Dict[str, Any]] = None,
        end_on_no_response=False,
        publish_proposer=True,
        publish_n_acceptances=False,
        enable_callbacks=False,
        avoid_ultimatum=False,
        check_offers=True,
        ignore_negotiator_exceptions=False,
        offering_is_accepting=True,
        name: Optional[str] = None,
        max_wait: int = sys.maxsize,
        **kwargs,
    ):
        """
        """
        super().__init__(
            issues=issues,
            outcomes=outcomes,
            n_steps=n_steps,
            time_limit=time_limit if time_limit is not None else float("inf"),
            step_time_limit=step_time_limit
            if step_time_limit is not None
            else float("inf"),
            max_n_agents=max_n_agents,
            dynamic_entry=dynamic_entry,
            keep_issue_names=keep_issue_names,
            outcome_type=outcome_type,
            cache_outcomes=cache_outcomes,
            max_n_outcomes=max_n_outcomes,
            annotation=annotation,
            state_factory=SAOState,
            enable_callbacks=enable_callbacks,
            name=name,
            **kwargs,
        )
        self.ignore_negotiator_exceptions = ignore_negotiator_exceptions
        self._current_offer = None
        self._current_proposer = None
        self._last_checked_negotiator = -1
        self._n_accepting = 0
        self._avoid_ultimatum = n_steps is not None and avoid_ultimatum
        self.end_negotiation_on_refusal_to_propose = end_on_no_response
        self.publish_proposer = publish_proposer
        self.publish_n_acceptances = publish_n_acceptances
        self.check_offers = check_offers
        self._no_responses = 0
        self._new_offers = []
        self._offering_is_accepting = offering_is_accepting
        self._n_waits = 10
        self._n_max_waits = max_wait if max_wait is not None else float("inf")

    def join(
        self,
        ami: AgentMechanismInterface,
        state: MechanismState,
        *,
        ufun: Optional["UtilityFunction"] = None,
        role: str = "agent",
    ) -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        if not self.ami.dynamic_entry and not any(
            [a.capabilities.get("propose", False) for a in self.negotiators]
        ):
            self._current_proposer = None
            self._last_checked_negotiator = -1
            self._current_offer = None
            self._n_accepting = 0
            return False
        return True

    def extra_state(self):
        current_proposer_agent = (
            self._current_proposer.owner if self._current_proposer else None
        )
        if current_proposer_agent and self.publish_proposer:
            current_proposer_agent = current_proposer_agent.id
        new_offerer_agents = []
        for neg_id, outcome in self._new_offers:
            neg = self._negotiator_map.get(neg_id, None)
            agent = neg.owner if neg else None
            if agent is not None and self.publish_proposer:
                new_offerer_agents.append(agent.id)
            else:
                new_offerer_agents.append(None)
        return dict(
            current_offer=self._current_offer,
            new_offers=self._new_offers,
            current_proposer=self._current_proposer.id
            if self._current_proposer and self.publish_proposer
            else None,
            current_proposer_agent=current_proposer_agent,
            n_acceptances=self._n_accepting if self.publish_n_acceptances else 0,
            new_offerer_agents=new_offerer_agents,
        )

    def plot(
        self,
        visible_negotiators: Union[Tuple[int, int], Tuple[str, str]] = (0, 1),
        plot_utils=True,
        plot_outcomes=False,
        utility_range: Optional[Tuple[float, float]] = None,
    ):
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec

        if self.issues is not None and len(self.issues) > 1:
            plot_outcomes = False

        if len(self.negotiators) < 2:
            print("Cannot visualize negotiations with more less than 2 negotiators")
            return
        if len(visible_negotiators) > 2:
            print("Cannot visualize more than 2 agents")
            return
        if isinstance(visible_negotiators[0], str):
            tmp = []
            for _ in visible_negotiators:
                for n in self.negotiators:
                    if n.id == _:
                        tmp.append(n)
        else:
            visible_negotiators = [
                self.negotiators[visible_negotiators[0]],
                self.negotiators[visible_negotiators[1]],
            ]
        indx = dict(zip([_.id for _ in self.negotiators], range(len(self.negotiators))))
        history = []
        for state in self.history:
            for a, o in state.new_offers:
                history.append(
                    {
                        "current_proposer": a,
                        "current_offer": o,
                        "offer_index": self.outcomes.index(o),
                        "relative_time": state.relative_time,
                        "step": state.step,
                        "u0": visible_negotiators[0].utility_function(o),
                        "u1": visible_negotiators[1].utility_function(o),
                    }
                )
        history = pd.DataFrame(data=history)
        has_history = len(history) > 0
        has_front = 1
        n_negotiators = len(self.negotiators)
        n_agents = len(visible_negotiators)
        ufuns = self._get_ufuns()
        outcomes = self.outcomes
        utils = [tuple(f(o) for f in ufuns) for o in outcomes]
        agent_names = [a.name for a in visible_negotiators]
        if has_history:
            history["offer_index"] = [outcomes.index(_) for _ in history.current_offer]
        frontier, frontier_outcome = self.pareto_frontier(sort_by_welfare=True)
        frontier_indices = [
            i
            for i, _ in enumerate(frontier)
            if _[0] is not None
            and _[0] > float("-inf")
            and _[1] is not None
            and _[1] > float("-inf")
        ]
        frontier = [frontier[i] for i in frontier_indices]
        frontier_outcome = [frontier_outcome[i] for i in frontier_indices]
        frontier_outcome_indices = [outcomes.index(_) for _ in frontier_outcome]
        if plot_utils:
            fig_util = plt.figure()
        if plot_outcomes:
            fig_outcome = plt.figure()
        gs_util = gridspec.GridSpec(n_agents, has_front + 1) if plot_utils else None
        gs_outcome = (
            gridspec.GridSpec(n_agents, has_front + 1) if plot_outcomes else None
        )
        axs_util, axs_outcome = [], []

        for a in range(n_agents):
            if a == 0:
                if plot_utils:
                    axs_util.append(fig_util.add_subplot(gs_util[a, has_front]))
                if plot_outcomes:
                    axs_outcome.append(
                        fig_outcome.add_subplot(gs_outcome[a, has_front])
                    )
            else:
                if plot_utils:
                    axs_util.append(
                        fig_util.add_subplot(gs_util[a, has_front], sharex=axs_util[0])
                    )
                if plot_outcomes:
                    axs_outcome.append(
                        fig_outcome.add_subplot(
                            gs_outcome[a, has_front], sharex=axs_outcome[0]
                        )
                    )
            if plot_utils:
                axs_util[-1].set_ylabel(agent_names[a])
            if plot_outcomes:
                axs_outcome[-1].set_ylabel(agent_names[a])
        for a, (au, ao) in enumerate(
            zip(
                itertools.chain(axs_util, itertools.repeat(None)),
                itertools.chain(axs_outcome, itertools.repeat(None)),
            )
        ):
            if au is None and ao is None:
                break
            if has_history:
                h = history.loc[
                    history.current_proposer == visible_negotiators[a].id,
                    ["relative_time", "offer_index", "current_offer"],
                ]
                h["utility"] = h["current_offer"].apply(ufuns[a])
                if plot_outcomes:
                    ao.plot(h.relative_time, h["offer_index"])
                if plot_utils:
                    au.plot(h.relative_time, h.utility)
                    if utility_range is not None:
                        au.set_ylim(*utility_range)

        if has_front:
            if plot_utils:
                axu = fig_util.add_subplot(gs_util[:, 0])
                axu.scatter(
                    [_[0] for _ in utils],
                    [_[1] for _ in utils],
                    label="outcomes",
                    color="gray",
                    marker="s",
                    s=20,
                )
            if plot_outcomes:
                axo = fig_outcome.add_subplot(gs_outcome[:, 0])
            clrs = ("blue", "green")
            if plot_utils:
                f1, f2 = [_[0] for _ in frontier], [_[1] for _ in frontier]
                axu.scatter(f1, f2, label="frontier", color="red", marker="x")
                # axu.legend()
                axu.set_xlabel(agent_names[0] + " utility")
                axu.set_ylabel(agent_names[1] + " utility")
                if self.agreement is not None:
                    pareto_distance = 1e9
                    cu = (ufuns[0](self.agreement), ufuns[1](self.agreement))
                    for pu in frontier:
                        dist = math.sqrt((pu[0] - cu[0]) ** 2 + (pu[1] - cu[1]) ** 2)
                        if dist < pareto_distance:
                            pareto_distance = dist
                    axu.text(
                        0.05,
                        0.05,
                        f"Pareto-distance={pareto_distance:5.2}",
                        verticalalignment="top",
                        transform=axu.transAxes,
                    )

            if plot_outcomes:
                axo.scatter(
                    frontier_outcome_indices,
                    frontier_outcome_indices,
                    color="red",
                    marker="x",
                    label="frontier",
                )
                axo.legend()
                axo.set_xlabel(agent_names[0])
                axo.set_ylabel(agent_names[1])

            if plot_utils and has_history:
                for a in range(n_agents):
                    h = history.loc[
                        history.current_proposer == visible_negotiators[a].id,
                        ["relative_time", "offer_index", "current_offer"],
                    ]
                    h["u0"] = h["current_offer"].apply(ufuns[0])
                    h["u1"] = h["current_offer"].apply(ufuns[1])
                    axu.scatter(h.u0, h.u1, color=clrs[a], label=f"{agent_names[a]}")
                axu.scatter(
                    [frontier[0][0]],
                    [frontier[0][1]],
                    color="magenta",
                    label=f"Max. Welfare",
                )
                axu.annotate(
                    "Max. Welfare",
                    xy=frontier[0],  # theta, radius
                    xytext=(
                        frontier[0][0] + 0.02,
                        frontier[0][1] + 0.02,
                    ),  # fraction, fraction
                    horizontalalignment="left",
                    verticalalignment="bottom",
                )
            if plot_outcomes and has_history:
                steps = sorted(history.step.unique().tolist())
                aoffers = [[], []]
                for step in steps[::2]:
                    offrs = []
                    for a in range(n_agents):
                        a_offer = history.loc[
                            (history.current_proposer == agent_names[a])
                            & ((history.step == step) | (history.step == step + 1)),
                            "offer_index",
                        ]
                        if len(a_offer) > 0:
                            offrs.append(a_offer.values[-1])
                    if len(offrs) == 2:
                        aoffers[0].append(offrs[0])
                        aoffers[1].append(offrs[1])
                axo.scatter(aoffers[0], aoffers[1], color=clrs[0], label=f"offers")

            if self.state.agreement is not None:
                if plot_utils:
                    axu.scatter(
                        [ufuns[0](self.state.agreement)],
                        [ufuns[1](self.state.agreement)],
                        color="black",
                        marker="*",
                        s=120,
                        label="SCMLAgreement",
                    )
                if plot_outcomes:
                    axo.scatter(
                        [outcomes.index(self.state.agreement)],
                        [outcomes.index(self.state.agreement)],
                        color="black",
                        marker="*",
                        s=120,
                        label="Agreement",
                    )

        if plot_utils:
            fig_util.show()
        if plot_outcomes:
            fig_outcome.show()

    def round(self) -> MechanismRoundResult:
        """implements a round of the Stacked Alternating Offers Protocol.


        """
        self._new_offers = []
        negotiators: List["SAONegotiator"] = self.negotiators
        n_negotiators = len(negotiators)

        def _safe_counter(negotiator, *args, **kwargs) -> SAOResponse:
            try:
                if (
                    negotiator == self._current_proposer
                ) and self._offering_is_accepting:
                    self._n_accepting = 0
                    response = negotiator.counter(*args, **kwargs)
                else:
                    response = negotiator.counter(*args, **kwargs)
            except Exception as ex:
                if self.ignore_negotiator_exceptions:
                    self.announce(
                        Event(
                            "negotiator_exception",
                            {"negotiator": negotiator, "exception": ex},
                        )
                    )
                    return SAOResponse(ResponseType.END_NEGOTIATION, None)
                else:
                    raise ex
            if (
                self.check_offers
                and response.outcome is not None
                and (not outcome_is_complete(response.outcome, self.issues))
            ):
                return SAOResponse(response.response, None)
            return response

        proposers, proposer_indices = [], []
        for i, neg in enumerate(negotiators):
            if not neg.capabilities.get("propose", False):
                continue
            proposers.append(neg)
            proposer_indices.append(i)
        n_proposers = len(proposers)
        if n_proposers < 1:
            if not self.dynamic_entry:
                return MechanismRoundResult(
                    broken=True,
                    timedout=False,
                    agreement=None,
                    error=True,
                    error_details="No proposers and no dynamic entry",
                )
            else:
                return MechanismRoundResult(
                    broken=False, timedout=False, agreement=None
                )
        # if this is the first step (or no one has offered yet) which means that there is no _current_offer
        if self._current_offer is None and n_proposers > 1 and self._avoid_ultimatum:
            if not self.dynamic_entry and not self.state.step == 0:
                if self.end_negotiation_on_refusal_to_propose:
                    return MechanismRoundResult(broken=True)
            # assert self._current_proposer is None
            # assert self._last_checked_negotiator == -1

            # if we are trying to avoid an ultimatum, we take an offer from everyone and ignore them but one.
            # this way, the agent cannot know its order. For example, if we have two agents and 3 steps, this will
            # be the situation after each step:
            #
            # Case 1: Assume that it ignored the offer from agent 1
            # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
            # 0   , counter(None)->offer1*  , counter(None) -> offer0   , 0/3
            # 1   , counter(offer2)->offer3 , counter(offer1) -> offer2 , 1/3
            # 2   , counter(offer4)->offer5 , counter(offer3) -> offer4 , 2/3
            # 3   ,                         , counter(offer5)->offer6   , 3/3
            #
            # Case 2: Assume that it ignored the offer from agent 0
            # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
            # 0   , counter(None)->offer1   , counter(None) -> offer0*  , 0/3
            # 1   , counter(offer0)->offer2 , counter(offer2) -> offer3 , 1/3
            # 2   , counter(offer3)->offer4 , counter(offer4) -> offer5 , 2/3
            # 3   , counter(offer5)->offer6 ,                           , 3/3
            #
            # in both cases, the agent cannot know whether its last offer going to be passed to the other agent
            # (the ultimatum scenario) or not.
            responses = []
            for neg in proposers:
                if not neg.capabilities.get("propose", False):
                    continue
                strt = time.perf_counter()
                resp = _safe_counter(neg, state=self.state, offer=None)
                if time.perf_counter() - strt > self.ami.step_time_limit:
                    return MechanismRoundResult(
                        broken=False, timedout=True, agreement=None
                    )
                if resp.response == ResponseType.END_NEGOTIATION:
                    return MechanismRoundResult(
                        broken=True, timedout=False, agreement=None
                    )
                if resp.response in (ResponseType.NO_RESPONSE, ResponseType.WAIT):
                    return MechanismRoundResult(
                        broken=True,
                        timedout=False,
                        agreement=None,
                        error=True,
                        error_details=f"When avoid_ultimatum is True, negotiators MUST return an offer when asked "
                        f"to propose the first time.\nNegotiator {neg.id} returned NO_RESPONSE/WAIT",
                    )
                responses.append(resp)
            if len(responses) < 1:
                if not self.dynamic_entry:
                    return MechanismRoundResult(
                        broken=True,
                        timedout=False,
                        agreement=None,
                        error=True,
                        error_details="No proposers and no dynamic entry",
                    )
                else:
                    return MechanismRoundResult(
                        broken=False, timedout=False, agreement=None
                    )
            # choose a random negotiator and set it as the current negotiator
            selected = random.randint(0, len(responses) - 1)
            resp = responses[selected]
            neg = proposers[selected]
            _first_proposer = proposer_indices[selected]
            self._n_accepting = 1 if self._offering_is_accepting else 0
            self._current_offer = resp.outcome
            self._current_proposer = neg
            self._last_checked_negotiator = _first_proposer
            self._new_offers.append((neg.id, resp.outcome))
            return MechanismRoundResult(broken=False, timedout=False, agreement=None)

        # this is not the first round. A round will get n_negotiators steps
        ordered_indices = [
            (_ + self._last_checked_negotiator + 1) % n_negotiators
            for _ in range(n_negotiators)
        ]
        for neg_indx in ordered_indices:
            self._last_checked_negotiator = neg_indx
            neg = self.negotiators[neg_indx]
            strt = time.perf_counter()
            resp = _safe_counter(neg, state=self.state, offer=self._current_offer)
            if time.perf_counter() - strt > self.ami.step_time_limit:
                return MechanismRoundResult(broken=False, timedout=True, agreement=None)
            if resp.response != ResponseType.WAIT:
                self._n_waits = 0
            if self._enable_callbacks:
                if self._current_offer is not None:
                    for other in self.negotiators:
                        if other is not neg:
                            other.on_partner_response(
                                state=self.state,
                                partner_id=neg.id,
                                outcome=self._current_offer,
                                response=resp.response,
                            )
            if resp.response == ResponseType.NO_RESPONSE:
                continue
            if resp.response == ResponseType.WAIT:
                self._n_waits += 1
                if self._n_waits > self._n_max_waits:
                    self._n_waits = 0
                    return MechanismRoundResult(
                        broken=False, timedout=True, agreement=None, waiting=False
                    )
                self._last_checked_negotiator = neg_indx - 1
                if neg_indx < 0:
                    self._last_checked_negotiator = n_negotiators - 1
                return MechanismRoundResult(
                    broken=False, timedout=False, agreement=None, waiting=True
                )
            if resp.response == ResponseType.END_NEGOTIATION:
                return MechanismRoundResult(broken=True, timedout=False, agreement=None)
            if resp.response == ResponseType.ACCEPT_OFFER:
                self._n_accepting += 1
                if self._n_accepting == n_negotiators:
                    return MechanismRoundResult(
                        broken=False, timedout=False, agreement=self._current_offer
                    )
            if resp.response == ResponseType.REJECT_OFFER:
                proposal = resp.outcome
                if proposal is None:
                    if (
                        neg.capabilities.get("propose", True)
                        and self.end_negotiation_on_refusal_to_propose
                    ):
                        return MechanismRoundResult(
                            broken=True, timedout=False, agreement=None
                        )
                    continue
                else:
                    self._current_offer = proposal
                    self._current_proposer = neg
                    self._new_offers.append((neg.id, proposal))
                    self._n_accepting = 1 if self._offering_is_accepting else 0
                    if self._enable_callbacks:
                        for other in self.negotiators:
                            if other is neg:
                                continue
                            other.on_partner_proposal(
                                partner_id=neg.id, offer=proposal, state=self.state
                            )
        return MechanismRoundResult(broken=False, timedout=False, agreement=None)

    def negotiator_offers(self, negotiator_id: str) -> List[Outcome]:
        offers = []
        for state in self._history:
            offers += [o for n, o in state.new_offers if n == negotiator_id]
        return offers


SAOProtocol = SAOMechanism
"""An alias for `SAOMechanism` object"""



"""Defines basic concept related to outcomes

Outcomes in this package are always assumed to be multi-issue outcomes where  single-issue outcomes can be implemented
as the special case with a single issue.

- Both Continuous and discrete issues are supported. All issue will have names. If none is given, a random name will be
  used. It is HIGHLY recommended to always name your issues.
- Outcomes are dictionaries with issue names as keys and issue values as values.

Examples:

  Different ways to create issues:

  >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')
  ...           , Issue(20, 'count')]
  >>> for _ in issues: print(_)
  price: (0.5, 2.0)
  date: ['2018.10.1', '2018.10.2', '2018.10.3']
  count: (0, 19)

  Outcome example compatible with the given set of issues:

  >>> a = {'price': 1.2, 'date': '2018.10.04', 'count': 4}

"""

import copy
import itertools
import numbers
import random
import warnings
import xml.etree.ElementTree as ET
from collections import defaultdict
from dataclasses import dataclass, fields
from enum import Enum
from functools import reduce
from operator import mul
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import numpy as np

from .common import NamedObject
from .generics import ienumerate, iget, ikeys, ivalues
from .helpers import unique_name
from .java import PYTHON_CLASS_IDENTIFIER

if TYPE_CHECKING:
    from negmas import Mechanism

LARGE_NUMBER = 100
__all__ = [
    "Outcome",
    "OutcomeType",
    "OutcomeRange",
    "ResponseType",
    "Issue",
    "Issues",
    "outcome_is_valid",
    "outcome_is_complete",
    "outcome_range_is_valid",
    "outcome_range_is_complete",
    "outcome_in_range",
    "enumerate_outcomes",
    "sample_outcomes",
    "outcome_as_dict",
    "outcome_as_tuple",
    "outcome_as",
    "is_outcome",
    "num_outcomes",
    "outcome_for",
]


class ResponseType(Enum):
    """Possible answers to offers during negotiation."""

    ACCEPT_OFFER = 0
    REJECT_OFFER = 1
    END_NEGOTIATION = 2
    NO_RESPONSE = 3
    WAIT = 4


class Issue(NamedObject):
    """Encodes an Issue.

    Args:
            values: Possible values for the issue
            name: Name of the issue. If not given, a random name will be generated
            min_value: Minimum value ( used only if the issue values parameter was a callable )
            max_value: Maximum value ( used only if the issue values parameter was a callable )
            value_type: If a type is given, all values will be forced to this type

    Examples:

        >>> print(Issue(['to be', 'not to be'], name='THE problem'))
        THE problem: ['to be', 'not to be']
        >>> print(Issue(3, name='Cars'))
        Cars: (0, 2)
        >>> print(Issue((0.0, 1.0), name='required accuracy'))
        required accuracy: (0.0, 1.0)
        >>> a = Issue((0.0, 1.0), name='required accuracy')
        >>> a.is_uncountable()
        True
        >>> a.is_countable()
        False
        >>> a.is_continuous()
        True
        >>> a = Issue(lambda: random.randint(10), min_value=0, max_value=9)
        >>> a.is_uncountable()
        True
        >>> a.is_countable()
        False
        >>> a.is_continuous()
        False

    Remarks:

        - Issues can be initialized by either an iterable of strings, an integer or a tuple of two values with
          the following meanings:

          - ``list of anything`` : This is an issue that can any value within the given set of values
            (strings, ints, floats, etc)
          - ``int`` : This is an issue that takes any value from 0 to the given value -1 (int)
          - Tuple[ ``float`` , ``float`` ] : This is an issue that can take any real value between the given limits (min, max)
          - Tuple[ ``int`` , ``int`` ] : This is an issue that can take any integer value between the given limits (min, max)
          - ``Callable`` : The callable should take no parameters and should act as a generator of issue values. This
             type of issue is always assumed to be neither countable nor continuous and are called uncountable. For
             example, you can use this type to make an issue that generates all integers from 0 to infinity.
        - If a list is given, min, max must be callable on it.
    """

    def __init__(
        self,
        values: Union[
            List[Any],
            int,
            Tuple[float, float],
            Tuple[int, int],
            List[int],
            Callable[[], Any],
        ],
        name: Optional[str] = None,
        min_value: Any = None,
        max_value: Any = None,
        value_type: Optional[Type] = None,
    ) -> None:
        super().__init__(name=name)

        if (
            value_type is not None
            and value_type is not numbers.Integral
            and isinstance(values, numbers.Integral)
        ):
            raise ValueError(
                "Cannot force a type that is not int while passing a singular integer as values"
            )
        # if isinstance(values, int) and values <= LARGE_NUMBER:
        #    values = list(range(values))
        self._value_type: Type = value_type
        self._n_values: int = float("inf")
        self._is_generator: bool = False
        self._is_range: bool = False
        self._is_int_range: bool = False
        self._is_float_range: bool = False
        self._is_integer_valued: bool = False
        self._is_real_valued: bool = False

        if isinstance(values, numbers.Integral):
            values = (0, int(values) - 1)

        if isinstance(values, tuple):
            if len(values) != 2:
                raise ValueError(
                    f"Passing {values} is illegal. Issues with ranges need 2-values tuples"
                )

            if value_type is not None:
                values = (value_type(values[0]), value_type(values[1]))
            isint0, isint1 = (
                isinstance(values[0], numbers.Integral),
                isinstance(values[1], numbers.Integral),
            )
            isreal0, isreal1 = (
                isinstance(values[0], numbers.Real)
                and not isinstance(values[0], numbers.Integral),
                isinstance(values[1], numbers.Real)
                and not isinstance(values[0], numbers.Integral),
            )

            if (
                (isint0 and not isint1)
                or (isint1 and not isint0)
                or (isreal0 and not isreal1)
                or (isreal1 and not isreal0)
                or (not (isint0 or isreal0) or not (isint1 or isreal1))
            ):
                raise ValueError(
                    f"Confusing types: Received types ({type(values[0])}, {type(values[1])})"
                )

            values = (values[0], values[1])
            self._is_int_range = isinstance(values[0], numbers.Integral)
            self._is_float_range = (
                isinstance(values[0], numbers.Real) and not self._is_int_range
            )
            self._is_range = True
            self._value_type = type(values[0])
            self._is_integer_valued = isinstance(values[0], numbers.Integral)
            self._is_real_valued = not self._is_integer_valued and isinstance(
                values[0], numbers.Real
            )

            if self._is_int_range:
                self._n_values = values[1] - values[0] + 1

        elif isinstance(values, Callable):
            self._is_generator = isinstance(values, Callable)
        else:
            values = list(values)

            if len(values) < 1:
                raise ValueError("values must include at least one item")

            if value_type is not None:
                values = [value_type(_) for _ in values]
            self._value_type = type(values[0])

            if value_type is None and not all(
                isinstance(_, self._value_type) for _ in values
            ):
                raise ValueError(f"Not all values are of the same type: {values}")
            self._n_values = len(values)
            self._is_integer_valued = isinstance(values[0], numbers.Integral)
            self._is_real_valued = not self._is_integer_valued and isinstance(
                values[0], numbers.Real
            )
        self._values = values

        if isinstance(self._values, tuple):
            self.min_value, self.max_value = values
        elif isinstance(self._values, list):
            self.min_value, self.max_value = min(values), max(values)
        else:
            self.min_value, self.max_value = min_value, max_value

    @property
    def values(self):
        return self._values

    @classmethod
    def from_outcomes(
        cls, outcomes: List["Outcome"], numeric_as_ranges: bool = False
    ) -> List["Issue"]:
        """
        Create a set of issues given some outcomes

        Args:

            outcomes: A list of outcomes
            numeric_as_ranges: If True, all numeric issues generated will have ranges that are defined by the minimum
                               and maximum values of that issue in the given outcomes instead of a list of the values
                               that appeared in them.

        Returns:

            a list of issues that include the given outcomes.

        Remarks:

            - The outcome space spanned by the generated issues can in principle contain many more possible outcomes
              than the ones given

        """

        def convert_type(v, old, values):
            if isinstance(v, numbers.Integral) and not isinstance(
                old, numbers.Integral
            ):
                return float(v)

            if not isinstance(v, numbers.Integral) and isinstance(
                old, numbers.Integral
            ):
                for i, _ in enumerate(values):
                    values[i] = float(_)

                return v

            if isinstance(v, str) and (isinstance(old, numbers.Number)):
                raise ValueError("a string after a number")

            if isinstance(old, str) and (isinstance(v, numbers.Number)):
                raise ValueError("a number after a string")

            return v

        names = None
        n_issues = None
        values = defaultdict(list)

        for i, o in enumerate(outcomes):
            o = outcome_as_dict(o)

            if n_issues is not None and len(o) != n_issues:
                raise ValueError(
                    f"Outcome {o} at {i} has {len(o)} issues but an earlier outcome had {n_issues} issues"
                )
            n_issues = len(o)

            if names is not None and not all(a == b for a, b in zip(names, o.keys())):
                raise ValueError(
                    f"Outcome {o} at {i} has issues {list(o.keys())} but an earlier outcome had issues {names}"
                )
            names = list(o.keys())

            for k, v in o.items():
                if len(values[k]) > 0:
                    try:
                        v = convert_type(v, values[k][-1], values[k])
                    except ValueError as e:
                        raise ValueError(
                            f"Outcome {o} at {i} has value {v} for issue {k} which is incompatible with an earlier "
                            f"value {values[k][-1]} ({str(e)})"
                        )
                values[k].append(v)

        for k, vals in values.items():
            values[k] = sorted(list(set(vals)))

        if numeric_as_ranges:
            return [
                Issue(values=(v[0], v[-1]), name=n)
                if len(v) > 0 and (isinstance(v[0], numbers.Number))
                else Issue(values=v, name=n)
                for n, v in values.items()
            ]
        else:
            return [Issue(values=v, name=n) for n, v in values.items()]

    @classmethod
    def to_xml_str(cls, issues: List["Issue"], enumerate_integer: bool = True) -> str:
        """Converts the list of issues into a well-formed xml string

        Examples:

            >>> issues = [Issue(values=10, name='i1'), Issue(values=['a', 'b', 'c'], name='i2'),
            ... Issue(values=(2.5, 3.5), name='i3')]
            >>> s = Issue.to_xml_str(issues)
            >>> print(s.strip())
            <negotiation_template>
            <utility_space number_of_issues="3">
            <objective description="" etype="objective" index="0" name="root" type="objective">
                <issue etype="discrete" index="1" name="i1" type="discrete" vtype="discrete">
                    <item index="1" value="0" cost="0" description="0">
                    </item>
                    <item index="2" value="1" cost="0" description="1">
                    </item>
                    <item index="3" value="2" cost="0" description="2">
                    </item>
                    <item index="4" value="3" cost="0" description="3">
                    </item>
                    <item index="5" value="4" cost="0" description="4">
                    </item>
                    <item index="6" value="5" cost="0" description="5">
                    </item>
                    <item index="7" value="6" cost="0" description="6">
                    </item>
                    <item index="8" value="7" cost="0" description="7">
                    </item>
                    <item index="9" value="8" cost="0" description="8">
                    </item>
                    <item index="10" value="9" cost="0" description="9">
                    </item>
                </issue>
                <issue etype="discrete" index="2" name="i2" type="discrete" vtype="discrete">
                    <item index="1" value="a" cost="0" description="a">
                    </item>
                    <item index="2" value="b" cost="0" description="b">
                    </item>
                    <item index="3" value="c" cost="0" description="c">
                    </item>
                </issue>
                <issue etype="real" index="3" name="i3" type="real" vtype="real">
                    <range lowerbound="2.5" upperbound="3.5"></range>
                </issue>
            </objective>
            </utility_space>
            </negotiation_template>

            >>> issues2, _ = Issue.from_xml_str(s)
            >>> print([_.__class__.__name__ for _ in issues2])
            ['Issue', 'Issue', 'Issue']
            >>> print(len(issues2))
            3
            >>> print([str(_) for _ in issues2])
            ["i1: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']", "i2: ['a', 'b', 'c']", 'i3: (2.5, 3.5)']
            >>> print([_.values for _ in issues2])
            [['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'], ['a', 'b', 'c'], (2.5, 3.5)]


        """
        output = (
            f'<negotiation_template>\n<utility_space number_of_issues="{len(issues)}">\n'
            f'<objective description="" etype="objective" index="0" name="root" type="objective">\n'
        )

        for indx, issue in enumerate(issues):
            if issue._values is None or issue._is_generator:
                raise ValueError(
                    f"Cannot convert issue {issue.name} to xml because it has no defined values (or uses a "
                    f"generator)."
                )

            if issue._is_int_range:
                if enumerate_integer:
                    output += f'    <issue etype="discrete" index="{indx + 1}" name="{issue.name}" type="discrete" vtype="discrete">\n'

                    for i, v in enumerate(
                        range(issue._values[0], issue._values[1] + 1)
                    ):
                        output += f'        <item index="{i + 1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
                    output += "    </issue>\n"
                else:
                    output += (
                        f'    <issue etype="integer" index="{indx + 1}" name="{issue.name}" type="integer" vtype="integer"'
                        f' lowerbound="{issue._values[0]}" upperbound="{issue._values[1]}" />\n'
                    )
            elif issue._is_float_range:
                output += (
                    f'    <issue etype="real" index="{indx + 1}" name="{issue.name}" type="real" vtype="real">\n'
                    f'        <range lowerbound="{issue._values[0]}" upperbound="{issue._values[1]}"></range>\n    </issue>\n'
                )
            else:
                output += f'    <issue etype="discrete" index="{indx + 1}" name="{issue.name}" type="discrete" vtype="discrete">\n'

                for i, v in enumerate(issue._values):
                    output += f'        <item index="{i + 1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
                output += "    </issue>\n"
        output += f"</objective>\n</utility_space>\n</negotiation_template>"

        return output

    @classmethod
    def to_genius(
        cls, issues: Iterable["Issue"], file_name: str, enumerate_integer: bool = True
    ) -> None:
        """Exports a the domain issues to a GENIUS XML file.

        Args:

            issues: The issues to be exported
            file_name (str): File name to export to

        Returns:

            A List[Issue] or Dict[Issue]


        Examples:

            >>> import pkg_resources
            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
            >>> Issue.to_genius(issues=issues, file_name = pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> issues2, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues2]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD

            - Forcing Single outcome

            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
            ...     , force_single_issue=True, keep_value_names=False, keep_issue_names=False)
            >>> print(list(issues[0].all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]
            >>> Issue.to_genius(issues=issues, enumerate_integer=True
            ...     , file_name = pkg_resources.resource_filename('negmas'
            ...                                   , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> issues3, _ = Issue.from_genius(file_name=pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> print([list(issue.all) for issue in issues3])
            [['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26']]

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        with open(file_name, "w") as f:
            f.write(
                cls.to_xml_str(issues=list(issues), enumerate_integer=enumerate_integer)
            )

    @classmethod
    def from_xml_str(
        cls,
        xml_str: str,
        force_single_issue=False,
        force_numeric=False,
        keep_value_names=True,
        keep_issue_names=True,
        safe_parsing=True,
        n_discretization: Optional[int] = None,
        max_n_outcomes: int = 1e6,
    ):
        """Exports a list/dict of issues from a GENIUS XML file.

        Args:

            xml_str (str): The string containing GENIUS style XML domain issue definitions
            force_single_issue (bool): Tries to generate a MappingUtility function with a single issue which is the
            product of all issues in the input
            keep_value_names (bool): Keep names of values
            keep_issue_names (bool): Keep names of issues
            safe_parsing (bool): Turn on extra checks
            n_discretization (Optional[int]): If not None, real valued issues are discretized with the given
            number of values
            max_n_outcomes (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

        Returns:

            - List[Issue] The issues (note that issue names will be stored in the name attribute of each issue if keep_issue_names)
            - List[dict] A list of agent information dicts each contains 'agent', 'class', 'utility_file_name'

        Examples:

            >>> import pkg_resources
            >>> domain_file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=False, keep_issue_names=False)
            >>> issue = issues[0]
            >>> print(issue.cardinality)
            27
            >>> print(list(issue.all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=False, keep_issue_names=True)
            >>> print(issues[0].name)
            Laptop-Harddisk-External Monitor
            >>> print(len(issues))
            1
            >>> print(list(issues[0].all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=True, keep_issue_names=False)
            >>> issue = issues[0]
            >>> print(issue.cardinality)
            27
            >>> print('\\n'.join(list(issue.all)[:5]))
            Dell+60 Gb+19'' LCD
            Dell+60 Gb+20'' LCD
            Dell+60 Gb+23'' LCD
            Dell+80 Gb+19'' LCD
            Dell+80 Gb+20'' LCD

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=True)
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0])
            "0: ['Dell', 'Macintosh', 'HP']"
            >>> print([_.cardinality for _ in issues])
            [3, 3, 3]
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=True, keep_value_names=True)
            >>> len(issues)
            3
            >>> str(issues[0])
            "Laptop: ['Dell', 'Macintosh', 'HP']"
            >>> print([_.cardinality for _ in issues])
            [3, 3, 3]
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD


            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=False)
            >>> len(issues)
            3
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0]).split(': ')[-1]
            '(0, 2)'
            >>> print([_.cardinality for _ in issues])
            [3, 3, 3]

            >>> domain_file_name = pkg_resources.resource_filename('negmas'
            ...                              , resource_name='tests/data/fuzzyagent/single_issue_domain.xml')
            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=False)
            >>> len(issues)
            1
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0]).split(': ')[-1]
            '(10.0, 40.0)'
            >>> print([_.cardinality for _ in issues])
            [inf]
        """
        root = ET.fromstring(xml_str)

        if safe_parsing and root.tag != "negotiation_template":
            raise ValueError(f"Root tag is {root.tag}: negotiation_template")

        utility_space = None
        agents = []

        for child in root:
            if child.tag == "utility_space":
                utility_space = child

                for _ in utility_space:
                    if _.tag == "objective":
                        utility_space = _

                        break
            elif child.tag == "agent":
                agents.append(child.attrib)

        if utility_space is None:
            if safe_parsing:
                raise ValueError(f"No objective child was found in the root")
            utility_space = root
        weights = {}
        issues = {}
        issue_info = {}
        all_discrete = True

        for child in utility_space:
            if child.tag == "issue":
                indx = int(child.attrib["index"]) - 1
                myname = child.attrib["name"]
                issue_key = myname if keep_issue_names else indx
                issue_info[issue_key] = {"name": myname, "index": indx}
                info = {"type": "discrete", "etype": "discrete", "vtype": "discrete"}

                for a in ("type", "etype", "vtype"):
                    info[a] = child.attrib.get(a, info[a])
                mytype = info["type"]

                if mytype == "discrete":
                    issues[issue_key] = []

                    for item in child:
                        if item.tag == "item":
                            item_indx = int(item.attrib["index"]) - 1
                            item_name = item.attrib.get("value", None)
                            item_key = (
                                item_name
                                if keep_value_names
                                and item_name is not None
                                and not force_numeric
                                else item_indx
                            )

                            if (
                                item_key not in issues[issue_key]
                            ):  # ignore repeated items
                                issues[issue_key].append(item_key)

                    if not keep_value_names:
                        issues[issue_key] = len(issues[issue_key])
                elif mytype in ("integer", "real"):
                    lower, upper = (
                        child.attrib.get("lowerbound", None),
                        child.attrib.get("upperbound", None),
                    )

                    for rng_child in child:
                        if rng_child.tag == "range":
                            lower, upper = (
                                rng_child.attrib.get("lowerbound", lower),
                                rng_child.attrib.get("upperbound", upper),
                            )

                    if mytype == "integer":
                        lower, upper = int(lower), int(upper)

                        if keep_value_names:
                            issues[issue_key] = list(range(lower, upper + 1))
                        else:
                            issues[issue_key] = upper - lower + 1
                    else:
                        if n_discretization is None:
                            all_discrete = False
                            issues[issue_key] = (float(lower), float(upper))
                        else:
                            issues[issue_key] = n_discretization
                else:
                    # I should add the real-valued issues code here
                    raise ValueError(f"Unknown type: {mytype}")
            else:
                raise ValueError(f"Unknown child for objective: {child.tag}")

        for key, value in zip(ikeys(issues), ivalues(issues)):
            issues[key] = Issue(
                values=value,
                name=issue_info[key]["name"]
                if keep_issue_names
                else str(issue_info[key]["index"]),
            )
        issues = list(issues.values())

        if force_single_issue:
            issue_name_ = (
                "-".join([_.name for _ in issues]) if keep_issue_names else "0"
            )

            if all_discrete or n_discretization is not None:
                n_outcomes = None

                if max_n_outcomes is not None:
                    n_items = [len(list(_.alli(n=n_discretization))) for _ in issues]
                    n_outcomes = reduce(mul, n_items, 1)

                    if n_outcomes > max_n_outcomes:
                        return None, None

                if keep_value_names:
                    if len(issues) > 1:
                        all_values = itertools.product(
                            *[
                                [str(_) for _ in issue.alli(n=n_discretization)]
                                for issue_key, issue in zip(range(len(issues)), issues)
                            ]
                        )
                        all_values = list(
                            map(lambda items: "+".join(items), all_values)
                        )
                    else:
                        all_values = [
                            str(_) for _ in issues[0].alli(n=n_discretization)
                        ]
                    issues = [Issue(values=all_values, name=issue_name_)]
                else:
                    if n_outcomes is None:
                        n_items = [_.cardinality for _ in issues]
                        n_outcomes = reduce(mul, n_items, 1)
                    issues = [Issue(values=n_outcomes, name=issue_name_)]
            else:
                return None, None

        return issues, agents

    @classmethod
    def from_genius(
        cls,
        file_name: str,
        force_single_issue=False,
        force_numeric=False,
        keep_value_names=True,
        keep_issue_names=True,
        safe_parsing=True,
        n_discretization: Optional[int] = None,
        max_n_outcomes: int = 1e6,
    ):
        """Imports a the domain issues from a GENIUS XML file.

        Args:

            max_n_outcomes: Maximum number of outcomes to use
            n_discretization: Number of discretization levels per issue
            safe_parsing: Add more checks to parsing
            keep_issue_names: Use dictionaries instead of tuples to represent outcomes
            keep_value_names: keep value names in case of strings
            force_numeric: Force the issue values to be numeric
            force_single_issue: Combine all issues into a single issue
            file_name (str): File name to import from

        Returns:

            A List[Issue] or Dict[Issue]


        Examples:

            >>> import pkg_resources
            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
            >>> print([_.name for _ in issues])
            ['Laptop', 'Harddisk', 'External Monitor']

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        with open(file_name, "r", encoding="utf-8") as f:
            xml_str = f.read()

            return cls.from_xml_str(
                xml_str=xml_str,
                force_single_issue=force_single_issue,
                keep_value_names=keep_value_names,
                keep_issue_names=keep_issue_names,
                safe_parsing=safe_parsing,
                n_discretization=n_discretization,
                max_n_outcomes=max_n_outcomes,
                force_numeric=force_numeric,
            )

    @staticmethod
    def num_outcomes(issues: Iterable["Issue"]) -> Union[int, float]:
        """Returns the total number of outcomes in a set of issues. `-1` indicates infinity"""
        n = 1

        for issue in issues:
            n *= issue.cardinality

        return n

    @staticmethod
    def generate(
        issues: Sequence[
            Union[int, List[str], Tuple[int, int], Callable, Tuple[float, float]]
        ],
        counts: Optional[Sequence[int]] = None,
        names: Optional[Sequence[str]] = None,
    ) -> List["Issue"]:
        """Generates a set of issues with given parameters. Each is optionally repeated

        Args:

            issues: The parameters of the issues
            counts: The number of times to repeat each of the `issues`
            names: The names to assign to the issues. If None, then string representations of integers
                   starting from zero will be used.

        Returns:
            List['Issue']: The list of issues with given conditions

        """
        one_each = counts is None
        int_names = names is None
        result = []
        nxt = 0

        for i, issue in enumerate(issues):
            count = 1 if one_each else counts[i]  # type: ignore

            for j in range(count):
                name = str(nxt) if int_names else names[i]  # type: ignore
                # if count > 1:
                #    name = name + f' {j}'
                nxt += 1
                result.append(Issue(values=issue, name=name))

        return result

    @property
    def type(self) -> str:
        """The type of the issue.

        Returns:
            str: either 'continuous', 'uncountable' or 'discrete'

        Remarks:

            - IF values is set to None or to a callable, the issue is treated as continuous

        """

        if self._is_float_range:
            return "continuous"

        return "uncountable" if self._is_generator else "discrete"

    def is_continuous(self) -> bool:
        """Test whether the issue is a continuous issue

        Returns:
            bool: uncountable (including continuous) or not

        """

        return self._is_float_range

    def is_uncountable(self) -> bool:
        """Test whether the issue has uncountable possible outcomes

        Returns:
            bool: uncountable (including continuous) or not

        """

        return self._is_float_range or self._is_generator

    def is_countable(self) -> bool:
        """Test whether the issue is a discrete issue

        Returns:
            bool: countable or not

        """

        return not self._is_float_range and not self._is_generator

    is_discrete = is_countable

    @property
    def all(self) -> Generator:
        """A generator that generates all possible values.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import Issue
            >>> list(Issue(5).all)
            [0, 1, 2, 3, 4]

        """

        if self.is_uncountable():
            raise ValueError(
                "Cannot return all possibilities of a continuous/uncountable issue"
            )

        if self._is_int_range:
            yield from range(self._values[0], self._values[1] + 1)
        else:
            yield from self._values  # type: ignore

    def alli(self, n: Optional[int] = 10) -> Generator:
        """A generator that generates all possible values or samples n values for real Issues.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import Issue
            >>> list(Issue(5).all)
            [0, 1, 2, 3, 4]

        """

        if self._is_float_range:
            if n is None:
                raise ValueError("Real valued issue with no discretization value")
            yield from np.linspace(
                self._values[0], self._values[1], num=n, endpoint=True
            ).tolist()
        elif self._is_generator:
            if n is None:
                raise ValueError("Real valued issue with no discretization value")
            yield from (self._values() for _ in range(n))
        elif self._is_int_range:
            yield from range(self._values[0], self._values[1] + 1)
        else:
            yield from self._values  # type: ignore

    @property
    def cardinality(self) -> Union[int, float]:
        """The number of possible outcomes for the issue. Returns infinity for continuous and uncountable spaces"""

        return self._n_values

    def rand(self) -> Union[int, float, str]:
        """Picks a random valid value."""

        if self._is_float_range:
            return (
                random.random() * (self._values[1] - self._values[0]) + self._values[0]
            )  # type: ignore

        if self._is_int_range:
            return random.randint(*self._values)

        if self._is_generator:
            return self._values()

        return random.choice(self._values)  # type: ignore

    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> Iterable["Outcome"]:
        """Picks a random valid value."""

        if self._is_int_range:
            if n > self._values[1] and not with_replacement:
                if fail_if_not_enough:
                    raise ValueError(
                        f"Cannot sample {n} outcomes out of {self._values} without replacement"
                    )
                else:
                    return [_ for _ in range(self._values[0], self._values[1] + 1)]

            if with_replacement:
                return np.random.randint(
                    low=self._values[0], high=self._values[1] + 1, size=n
                ).tolist()
            else:
                return random.shuffle(
                    [_ for _ in range(self._values[0], self._values[1] + 1)]
                )[:n]

        if self._is_float_range:
            if with_replacement:
                return (
                    np.random.rand(n) * (self._values[1] - self._values[0])
                    + self._values[0]
                ).tolist()
            else:
                return np.linspace(
                    self._values[0], self._values[1], num=n, endpoint=True
                ).tolist()

        if self._is_generator:
            if not with_replacement:
                raise ValueError(
                    f"values is specified as a callables for this issue. Cannot sample from it without "
                    f"replacement"
                )

            return [self._values() for _ in range(n)]

        if n > len(self._values) and not with_replacement:
            if fail_if_not_enough:
                raise ValueError(
                    f"Cannot sample {n} outcomes out of {self._values} without replacement"
                )
            else:
                return self._values

        return np.random.choice(
            np.asarray(self._values, dtype=self._value_type),
            size=n,
            replace=with_replacement,
        ).tolist()

    rand_valid = rand

    def rand_invalid(self) -> Union[int, float, str]:
        """Pick a random *invalid* value"""

        if self._is_int_range:
            return random.randint(self.max_value + 1, 2 * self.max_value)

        if self._is_float_range:
            return (
                random.random() * (self.max_value - self.min_value)
                + self.max_value * 1.1
            )

        if self._is_generator:
            raise ValueError(
                f"Cannot generate invalid outcomes because values is given as a callable"
            )

        if self._is_real_valued:
            return random.random() * self.max_value + self.max_value * 1.1

        if self._is_integer_valued:
            return random.randint(self.max_value + 1, self.max_value * 2)

        return unique_name("") + str(random.choice(self._values)) + unique_name("")

    @classmethod
    def enumerate(
        cls,
        issues: Collection["Issue"],
        max_n_outcomes: int = None,
        astype: Type = dict,
    ) -> List["Outcome"]:
        """
        Enumerates the outcomes of a list of issues.

        Args:
            issues: The list of issues.
            max_n_outcomes: The maximum number of outcomes to return
            astype: The type to use for the returned outcomes. Can be `typle`
                    `dict`, or any `OutcomeType`.
        Returns:
            List of outcomes of the given type.
        """
        n = num_outcomes(issues)

        if n is None or n == float("inf"):
            return cls.sample(
                issues=issues,
                n_outcomes=max_n_outcomes,
                astype=astype,
                fail_if_not_enough=False,
                with_replacement=False,
            )
        values = enumerate_outcomes(issues, astype=tuple)

        if max_n_outcomes is not None and n > max_n_outcomes:
            values = random.sample(values, max_n_outcomes)

        if issubclass(astype, tuple):
            return values
        issue_names = [_.name for _ in issues]
        if issubclass(astype, dict):
            return [outcome_as_dict(value, issue_names) for value in values]
        return [astype(**outcome_as_dict(value, issue_names)) for value in values]

    @classmethod
    def discretize_and_enumerate(
        cls,
        issues: Collection["Issue"],
        n_discretization: int = 10,
        astype: Type = dict,
    ) -> List["Outcome"]:
        """
        Enumerates the outcomes of a list of issues.

        Args:
            issues: The list of issues.
            max_n_outcomes: The maximum number of outcomes to return
            astype: The type to use for the returned outcomes. Can be `typle`
                    `dict`, or any `OutcomeType`.
        Returns:
            List of outcomes of the given type.
        """
        issues = [
            _
            if _.is_countable()
            else Issue(values=_.alli(n_discretization), name=_.name)
            for _ in issues
        ]
        return cls.enumerate(issues, max_n_outcomes=None, astype=astype)

    @classmethod
    def sample(
        cls,
        issues: Collection["Issue"],
        n_outcomes: int,
        astype: Type = dict,
        with_replacement: bool = True,
        fail_if_not_enough=True,
    ) -> List["Outcome"]:
        """
        Samples some outcomes from the issue space defined by the list of issues

        Args:

            issues: List of issues to sample from
            n_outcomes: The number of outcomes required
            astype: The type of the outcomes. It can be `tuple`, `dict` or any `OutcomeType`
            with_replacement: Whether sampling is with replacement (allowing repetition)
            fail_if_not_enough: IF given then an exception is raised if not enough outcomes are available

        Returns:

            a list of outcomes

        Examples:

            >>> from negmas.outcomes import Issue, OutcomeType
            >>> issues = [Issue(name='price', values=(0.0, 3.0)), Issue(name='quantity', values=10)]

            Sampling outcomes as tuples

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=tuple)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == tuple
            True

            Sampling outcomes as dicts

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=dict)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == dict
            True
            >>> list(samples[0].keys())
            ['price', 'quantity']

            >>> @dataclass
            ... class MyOutcome(OutcomeType):
            ...     price: float = 0.0
            ...     quantity: int = 0


            Sampling outcomes as an arbitrary class

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=MyOutcome)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == MyOutcome
            True
            >>> list(samples[0].keys())
            ['price', 'quantity']


        """
        n_total = num_outcomes(issues)

        if (
            n_total is not None
            and n_total != float("inf")
            and n_outcomes is not None
            and n_total < n_outcomes
            and fail_if_not_enough
            and not with_replacement
        ):
            raise ValueError(
                f"Cannot sample {n_outcomes} from a total of possible {n_total} outcomes"
            )

        if n_total is not None and n_total != float("inf") and n_outcomes is None:
            values = enumerate_outcomes(issues=issues, astype=tuple)
        elif n_total is None and n_outcomes is None:
            raise ValueError(
                f"Cannot sample unknown number of outcomes from continuous outcome spaces"
            )
        else:
            samples = []

            for issue in issues:
                samples.append(
                    issue.rand_outcomes(
                        n=n_outcomes, with_replacement=True, fail_if_not_enough=True
                    )
                )
            values = []

            for i in range(n_outcomes):
                values.append([s[i] for s in samples])

            if not with_replacement and not any(i.is_uncountable() for i in issues):
                tmp_values = []

                for value in values:
                    tmp_values.append(tuple(value))
                tmp_values = set(tmp_values)
                remaining = n_outcomes - len(tmp_values)
                n_max_trials, i = 10, 0

                while remaining < 0 and i < n_max_trials:
                    tmp_values = tmp_values.union(
                        set(
                            cls.sample(
                                issues=issues,
                                n_outcomes=remaining,
                                astype=tuple,
                                with_replacement=True,
                                fail_if_not_enough=False,
                            )
                        )
                    )
                    i += 1
                values = list(tmp_values)

        if issubclass(astype, tuple):
            return [tuple(_) for _ in values]
        issue_names = [_.name for _ in issues]
        if issubclass(astype, dict):
            return [outcome_as_dict(value, issue_names) for value in values]
        return [astype(**outcome_as_dict(value, issue_names)) for value in values]

    @property
    def outcome_range(self) -> "OutcomeRange":
        """An outcome range that represents the full space of the issues"""
        outcome_range = {}

        if self._is_range:
            outcome_range[self.name] = self._values
        elif self._is_generator:
            outcome_range[self.name] = None
        else:
            outcome_range[self.name] = self.all

        return outcome_range

    def __str__(self):
        return f"{self.name}: {self._values}"

    __repr__ = __str__

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return self._values == other.values and self.name == other.name

    def __copy__(self):
        return Issue(name=self.name, values=self._values)

    def __deepcopy__(self, memodict={}):
        if isinstance(self._values, list):
            return Issue(name=self.name, values=[_ for _ in self._values])

        return Issue(name=self.name, values=self._values)

    @classmethod
    def from_java(cls, d: Dict[str, Any], class_name: str) -> "Issue":
        if class_name.endswith("ListIssue"):
            return Issue(name=d.get("name", None), values=d["values"])

        if class_name.endswith("RangeIssue"):
            return Issue(name=d.get("name", None), values=(d["min"], d["max"]))
        raise ValueError(
            f"Unknown issue type: {class_name} with dict {d} received from Java"
        )

    def to_java(self):
        if self._values is None:
            return None

        if self._is_int_range:
            return {
                "name": self.name,
                "min": int(self._values[0]),
                "max": int(self._values[0]),
                PYTHON_CLASS_IDENTIFIER: "negmas.outcomes.IntRangeIssue",
            }

        if self._is_float_range:
            return {
                "name": self.name,
                "min": float(self._values[0]),
                "max": float(self._values[0]),
                PYTHON_CLASS_IDENTIFIER: "negmas.outcomes.DoubleRangeIssue",
            }

        if self._is_generator:
            raise ValueError("Cannot convert issues created by a callable to JAVA")

        if self._is_integer_valued:
            return {
                "name": self.name,
                "values": [int(_) for _ in self._values],
                PYTHON_CLASS_IDENTIFIER: "negmas.outcomes.IntListIssue",
            }

        if self._is_real_valued:
            return {
                "name": self.name,
                "values": [float(_) for _ in self._values],
                PYTHON_CLASS_IDENTIFIER: "negmas.outcomes.DoubleListIssue",
            }

        return {
            "name": self.name,
            "values": [str(_) for _ in self._values],
            PYTHON_CLASS_IDENTIFIER: "negmas.outcomes.StringListIssue",
        }


class Issues(object):
    """Encodes a set of Issues.

    Args:

        name-value pairs



    Remarks:

        - Issues can be initialized by either an iterable of strings, an integer or a tuple of two real values with
          the following meanings:
          - ``iterable of strings``: This is an issue that can any value within the given set of values (strings)
          - ``int``: This is an issue that takes any value from 0 to the given value -1 (int)
          - ``float``: This is an issue that can take any real value between the given limits (min, max)

    """

    def __init__(self, **kwargs) -> None:
        self.issues = [Issue(name=k, values=v) for k, v in kwargs.items()]

    @classmethod
    def from_issue_collection(cls, issues: Iterable[Issue]):
        out = Issues()
        out.issues = issues

        return out

    @classmethod
    def from_single_issue(cls, issue: Issue):
        return Issues.from_issue_collection([issue])

    @property
    def num_outcomes(self) -> Union[int, float]:
        """Returns the total number of outcomes in a set of issues. Infinity is returned for uncountable or continuous
        outcomes """
        n = 1

        for issue in self.issues:
            n *= issue._n_values

        return n

    @property
    def types(self) -> List[str]:
        """The type of the issue.

        Returns:
            str: either 'continuous' or 'discrete'

        """
        types_ = []

        for issue in self.issues:
            types_.append(issue.type)

        return types_

    def is_infinite(self) -> bool:
        """Test whether any issue is continuous (infinite outcome space)

        Returns:
            bool: continuous or not

        """

        return any(_.startswith("c") for _ in self.types)

    def is_finite(self) -> bool:
        """Test whether all issues are discrete (finite outcome space)

        Returns:
            bool: discrete or not

        """

        return all(_.startswith("d") for _ in self.types)

    @property
    def all(self) -> Generator:
        """A generator that generates all possible values.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:


        """

        if self.is_infinite():
            raise ValueError("Cannot return all possibilities of a continuous issues")

        yield from itertools.product(_.all for _ in self.issues)

    cardinality = num_outcomes
    """The number of possible outcomes for the set of issues. """

    def rand(self) -> Dict[str, Union[int, float, str]]:
        """Picks a random valid value."""

        return {_.name: _.rand() for _ in self.issues}

    rand_valid = rand

    def rand_invalid(self) -> Dict[str, Union[int, float, str]]:
        """Pick a random *invalid* value"""

        return {_.name: _.rand_invalid() for _ in self.issues}

    @property
    def outcome_range(self) -> "OutcomeRange":
        """An outcome range that represents the full space of the issues"""
        outcome_range = {}

        for issue in self.issues:
            if issue.is_continuous():
                outcome_range[issue.name] = issue.values
            elif issue.is_uncountable():
                outcome_range[issue.name] = None
            else:
                outcome_range[issue.name] = issue.all

        return outcome_range

    def __str__(self):
        return "\n".join([str(_) for _ in self.issues])


@dataclass
class OutcomeType:
    """A helper class allowing for definition of types that behave as outcomes (either in the form of dict or tuple).

    This class is intended to be used when a simple tuple or dict is not enough for describing an outcome (e.g. to use
    editor features like auto-completion of members). You simply define your class as a dataclass and add your fields to
    it then inherit from OutcomeType. As we do nothing in the __init__ function, that is compatible with python
    dataclasses.


    Examples:

        >>> from negmas.outcomes import OutcomeType, Issue
        >>> @dataclass
        ... class MyOutcome(OutcomeType):
        ...     price: float = 0.0
        ...     quantity: int = 0

        You can use MyOutcome as an outcome directly or convert it to a tuple/dict for other functions

        >>> outcome = MyOutcome(price=2.0, quantity=3)
        >>> outcome.price
        2.0
        >>> outcome['price']
        2.0
        >>> outcome.astuple()
        (2.0, 3)
        >>> outcome.asdict()
        {'price': 2.0, 'quantity': 3}

        You can also use outputs from issues to initialize your class

        >>> issues = [Issue(name='price', values=(0.0, 3.0)), Issue(name='quantity', values=10)]
        >>> sample = Issue.sample(issues=issues, n_outcomes=1)[0]

        >>> outcome = MyOutcome(**sample)
        >>> outcome.price == outcome['price']
        True


    """

    def __getitem__(self, item):
        """Makes the outcome type behave like a dict"""

        return self.__dict__[item]

    def keys(self) -> List[str]:
        return tuple(_.name for _ in fields(self))

    def values(self) -> List[str]:
        return tuple(self.__dict__[_.name] for _ in fields(self))

    def astuple(self):
        """Converts the outcome to a tuple where the order of items is the same as they are defined as fields"""

        return tuple(self.__dict__[_.name] for _ in fields(self))

    def asdict(self):
        """Converts the outcome to a dict containing all fields"""

        return {_.name: self.__dict__[_.name] for _ in fields(self)}

    def get(self, name, default: Any = None):
        """Acts like dict.get"""
        try:
            return getattr(self, name, default)
        except:
            return default


Outcome = Union[
    OutcomeType,
    Tuple[Union[int, float, str, list]],
    Dict[Union[int, str], Union[int, float, str, list]],
]
"""An outcome is either a tuple of values or a dict with name/value pairs."""

Outcomes = List["Outcome"]
OutcomeRanges = List["OutcomeRange"]

OutcomeRange = Mapping[
    Union[int, str],
    Union[
        int,
        float,
        str,
        List[int],
        List[float],
        List[str],
        Tuple[int, int],
        Tuple[float, float],
        List[Tuple[Union[int, float], Union[int, float]]],
    ],
]
"""Represents a range of outcomes."""


def num_outcomes(issues: Collection[Issue]) -> Optional[int]:
    """
    Returns the total number of outcomes in a set of issues.
    `-1` indicates infinity
    """
    return Issue.num_outcomes(issues)


def is_outcome(x: Any) -> bool:
    """Checks if x is acceptable as an outcome type"""

    return isinstance(x, dict) or isinstance(x, tuple) or isinstance(x, OutcomeType)


def enumerate_outcomes(
    issues: Iterable[Issue], keep_issue_names=None, astype=dict
) -> Optional[Union[List["Outcome"], Dict[str, "Outcome"]]]:
    """Enumerates all outcomes of this set of issues if possible

    Args:
        issues: A list of issues
        keep_issue_names: DEPRECTED. use `astype` instead
        astype: The type to use for returning outcomes. Can be tuple, dict or any `OutcomeType`

    Returns:
        list of outcomes
    """
    if keep_issue_names is not None:
        warnings.warn(
            "keep_issue_names is depricated. Use outcome_type instead.\n"
            "keep_issue_names=True <--> outcome_type=dict\n"
            "keep_issue_names=False <--> outcome_type=tuple\n",
            DeprecationWarning,
        )
        astype = dict if keep_issue_names else tuple
    try:
        outcomes = list(tuple(_) for _ in itertools.product(*[_.all for _ in issues]))
    except:
        return None

    if issubclass(astype, dict):
        issue_names = [_.name for _ in issues]
        outcomes = [outcome_as_dict(_, issue_names) for _ in outcomes]
    elif not issubclass(astype, tuple):
        issue_names = [_.name for _ in issues]
        outcomes = [astype(**outcome_as_dict(_, issue_names)) for _ in outcomes]

    return outcomes


def sample_outcomes(
    issues: Iterable[Issue],
    n_outcomes: Optional[int] = None,
    keep_issue_names=None,
    astype=dict,
    min_per_dim=5,
    expansion_policy=None,
) -> Optional[List[Optional["Outcome"]]]:
    """Discretizes the issue space and returns either a predefined number of outcomes or uniform samples

    Args:
        issues: The issues describing the issue space to be discretized
        n_outcomes: If None then exactly `min_per_dim` bins will be used for every continuous dimension and all outcomes
        will be returned
        keep_issue_names: DEPRICATED. Use `astype` instead
        min_per_dim: Max levels of discretization per dimension
        expansion_policy: None or 'repeat' or 'null' or 'no'. If repeat, then some of the outcomes will be repeated
        if None or 'no' then no expansion will happen if the total number of outcomes is less than
        n_outcomes: If 'null' then expansion will be with None values
        astype: The type used for returning outcomes. Can be tuple, dict or any `OutcomeType`

    Returns:
        List of outcomes

    Examples:

        enumberate the whole space

        >>> issues = [Issue(values=(0.0, 1.0), name='Price'), Issue(values=['a', 'b'], name='Name')]
        >>> sample_outcomes(issues=issues)
        [{'Price': 0.0, 'Name': 'a'}, {'Price': 0.0, 'Name': 'b'}, {'Price': 0.25, 'Name': 'a'}, {'Price': 0.25, 'Name': 'b'}, {'Price': 0.5, 'Name': 'a'}, {'Price': 0.5, 'Name': 'b'}, {'Price': 0.75, 'Name': 'a'}, {'Price': 0.75, 'Name': 'b'}, {'Price': 1.0, 'Name': 'a'}, {'Price': 1.0, 'Name': 'b'}]

        enumerate with sampling for very large space (we have 10 outcomes in the discretized space)

        >>> issues = [Issue(values=(0, 1), name='Price', value_type=float), Issue(values=['a', 'b'], name='Name')]
        >>> issues[0].is_continuous()
        True
        >>> sampled=sample_outcomes(issues=issues, n_outcomes=5)
        >>> len(sampled)
        5
        >>> len(set(tuple(_.values()) for _ in sampled))
        5

        >>> issues = [Issue(values=(0, 1), name='Price'), Issue(values=['a', 'b'], name='Name')]
        >>> issues[0].is_continuous()
        False
        >>> sampled=sample_outcomes(issues=issues, n_outcomes=5)
        >>> len(sampled)
        4
        >>> len(set(tuple(_.values()) for _ in sampled))
        4

    """
    if keep_issue_names is not None:
        warnings.warn(
            "keep_issue_names is depricated. Use outcome_type instead.\n"
            "keep_issue_names=True <--> outcome_type=dict\n"
            "keep_issue_names=False <--> outcome_type=tuple\n",
            DeprecationWarning,
        )
        astype = dict if keep_issue_names else tuple
    issues = [copy.deepcopy(_) for _ in issues]
    continuous = []
    uncountable = []
    indx = []
    uindx = []
    discrete = []
    n_disc = 0

    for i, issue in enumerate(issues):
        if issue.is_continuous():
            continuous.append(issue)
            indx.append(i)
        elif issue.is_uncountable():
            uncountable.append(issue)
            uindx.append(i)
        else:
            discrete.append(issue)
            n_disc += issue.cardinality

    if len(continuous) > 0:
        if n_outcomes is not None:
            n_per_issue = max(min_per_dim, (n_outcomes - n_disc) / len(continuous))
        else:
            n_per_issue = min_per_dim

        for i, issue in enumerate(continuous):
            issues[indx[i]] = Issue(
                name=issue.name,
                values=list(
                    np.linspace(
                        issue.min_value, issue.max_value, num=n_per_issue, endpoint=True
                    ).tolist()
                ),
            )

    if len(uncountable) > 0:
        if n_outcomes is not None:
            n_per_issue = max(min_per_dim, (n_outcomes - n_disc) / len(uncountable))
        else:
            n_per_issue = min_per_dim

        for i, issue in enumerate(uncountable):
            issues[uindx[i]] = Issue(
                name=issue.name, values=[issue.values() for _ in range(n_per_issue)]
            )

    cardinality = 1

    for issue in issues:
        cardinality *= issue.cardinality

    if cardinality == n_outcomes or n_outcomes is None:
        return list(enumerate_outcomes(issues, astype=astype))

    if cardinality < n_outcomes:
        outcomes = list(enumerate_outcomes(issues, astype=astype))

        if expansion_policy == "no" or expansion_policy is None:
            return outcomes
        elif expansion_policy == "null":
            return outcomes + [None] * (n_outcomes - cardinality)
        elif expansion_policy == "repeat":
            n_reps = n_outcomes // cardinality
            n_rem = n_outcomes % cardinality

            if n_reps > 1:
                for _ in n_reps:
                    outcomes += outcomes

            if n_rem > 0:
                outcomes += outcomes[:n_rem]

            return outcomes

    return list(random.sample(enumerate_outcomes(issues, astype=astype), n_outcomes))


def _is_single(x):
    """Checks whether a value is a single value which is defined as either a string or not an Iterable."""

    return isinstance(x, str) or isinstance(x, numbers.Number)


def outcome_is_valid(outcome: "Outcome", issues: Collection[Issue]) -> bool:
    """Test validity of an outcome given a set of issues.

    Examples:
        >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , Issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: (0, 19)
        >>> print([outcome_is_valid({'price':3.0}, issues), outcome_is_valid({'date': '2018.10.4'}, issues)\
            , outcome_is_valid({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_valid(valid_incomplete, issues))
        True
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True

    Args:
        outcome: outcome tested which can contain values for a partial set of issue values
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure
    """
    outcome = outcome_as_dict(outcome, [_.name for _ in issues])

    for issue in issues:
        for key in ikeys(outcome):
            if str(issue.name) == str(key):
                break

        else:
            continue

        value = iget(outcome, key)

        if issue._is_range and (
            isinstance(value, str) or not issue.min_value <= value <= issue.max_value
        ):
            return False

        if isinstance(issue._values, list) and value not in issue._values:
            return False

    return True


def outcome_is_complete(outcome: "Outcome", issues: Collection[Issue]) -> bool:
    """Tests that the outcome is valid and complete.

    Examples:
        >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , Issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: (0, 19)
        >>> print([outcome_is_complete({'price':3.0}, issues), outcome_is_complete({'date': '2018.10.4'}, issues)\
            , outcome_is_complete({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True
        >>> invalid = {'price': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False
        >>> invalid = {'unknown': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False

     Args:
        outcome: outcome tested which much contain valid values all issues if it is to be considered complete.
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure

    """
    outcome = outcome_as_dict(outcome, [_.name for _ in issues])

    if len(outcome) != len(issues):
        return False

    valid = outcome_is_valid(outcome, issues)

    if not valid:
        return False

    outcome_keys = [str(k) for k in ikeys(outcome)]

    for issue in issues:
        if str(issue.name) not in outcome_keys:
            return False

    return True


def outcome_range_is_valid(
    outcome_range: OutcomeRange, issues: Optional[Collection[Issue]] = None
) -> Union[bool, Tuple[bool, str]]:
    """Tests whether the outcome range is valid for the set of issues.

    Args:
        outcome_range:
        issues:

    Example:
        >>> try:
        ...     outcome_range_is_valid({'price': (0, 10)})
        ... except NotImplementedError:
        ...     print('Not implemented')
        Not implemented

    Returns:

    """
    # TODO implement this function
    raise NotImplementedError()


def outcome_range_is_complete(
    outcome_range: OutcomeRange, issues: Optional[Collection[Issue]] = None
) -> Union[bool, Tuple[bool, str]]:
    """Tests whether the outcome range is valid and complete for the set of issues

    Args:
        outcome_range:
        issues:

    Example:
        >>> try:
        ...     outcome_range_is_complete({'price': (0, 10)})
        ... except NotImplementedError:
        ...     print('Not implemented')
        Not implemented

    Returns:

    """
    # TODO implement this function
    raise NotImplementedError()


#################################
# Outcome space implementation  #
#################################
def outcome_in_range(
    outcome: "Outcome",
    outcome_range: OutcomeRange,
    *,
    strict=False,
    fail_incomplete=False,
) -> bool:
    """
    Tests that the outcome is contained within the given range of outcomes.

    An outcome range defines a value or a range of values for each issue.

    Args:

        outcome: "Outcome" being tested
        outcome_range: "Outcome" range being tested against
        strict: Whether to enforce that all issues in the outcome must be mentioned in the outcome_range
        fail_incomplete: If True then outcomes that do not sepcify a value for all keys in the outcome_range
        will be considered not falling within it. If False then these outcomes will be considered falling
        within the range given that the values for the issues mentioned in the outcome satisfy the range
        constraints.

    Examples:

        >>> outcome_range = {'price': (0.0, 2.0), 'distance': [0.3, 0.4], 'type': ['a', 'b'], 'area': 3}
        >>> outcome_range_2 = {'price': [(0.0, 1.0), (1.5, 2.0)], 'area': [(3, 4), (7, 9)]}
        >>> outcome_in_range({'price':3.0}, outcome_range)
        False
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range, strict=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range, fail_incomplete=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range)
        True
        >>> outcome_in_range({'type': 'c'}, outcome_range)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3.1}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3}, outcome_range_2)
        False
        >>> outcome_in_range({'area': 5}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4}, outcome_range_2, fail_incomplete=True)
        False
        >>> outcome_in_range({'price': 1.2}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4, 'area': 3.9}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 4}, outcome_range_2)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range_2)
        True
        >>> outcome_in_range({'type': 'a'}, outcome_range_2, strict=True)
        False
        >>> outcome_range = {'price': 10}
        >>> outcome_in_range({'price': 10}, outcome_range)
        True
        >>> outcome_in_range({'price': 11}, outcome_range)
        False

    Returns:

        bool: Success or failure

    Remarks:
        Outcome ranges specify regions in an outcome space. They can have any of the following conditions:

        - A key/issue not mentioned in the outcome range does not add any constraints meaning that **All**
          values are acceptable except if strict == True. If strict == True then *NO* value will be accepted for issues
          not in the outcome_range.
        - A key/issue with the value None in the outcome range means **All** values on this issue are acceptable.
          This is the same as having this key/issue removed from the outcome space
        - A key/issue withe the value [] (empty list) accepts *NO* outcomes
        - A key/issue with  a single value means that it is the only one acceptable
        - A key/issue with a single 2-items tuple (min, max) means that any value within that range is acceptable.
        - A key/issue with a list of values means an output is acceptable if it falls within the condition specified
          by any of the values in the list (list == union). Each such value can be a single value, a 2-items
          tuple or another list. Notice that lists of lists can always be combined into a single list of values

    """

    if (
        fail_incomplete
        and len(set(ikeys(outcome_range)).difference(ikeys(outcome))) > 0
    ):
        return False

    for key, value in ienumerate(outcome):
        if key not in ikeys(outcome_range):
            if strict:
                return False

            continue

        values = iget(outcome_range, key, None)

        if values is None:
            return False

        if _is_single(values) and value != values:
            return False

        if isinstance(values, tuple) and not values[0] < value < values[1]:
            return False

        if isinstance(values, list):
            for constraint in values:
                if _is_single(constraint):
                    if value == constraint:
                        break

                elif isinstance(constraint, list):
                    if value in constraint:
                        break

                elif isinstance(constraint, tuple):
                    if constraint[0] < value < constraint[1]:
                        break

            else:
                return False

            continue

    return True


def outcome_as_dict(outcome: "Outcome", issue_names: List[str] = None):
    """Converts the outcome to a dict no matter what was its type"""

    if outcome is None:
        return None

    if isinstance(outcome, np.ndarray):
        outcome = tuple(outcome.tolist())

    if isinstance(outcome, dict):
        return outcome

    if isinstance(outcome, OutcomeType):
        return outcome.asdict()

    if issue_names is not None:
        return dict(zip(issue_names, outcome))

    return dict(zip((str(_) for _ in range(len(outcome))), outcome))


def outcome_as_tuple(outcome: "Outcome"):
    """Converts the outcome to a tuple no matter what was its type"""

    if outcome is None:
        return None

    if isinstance(outcome, tuple):
        return outcome

    if isinstance(outcome, dict):
        return tuple(list(outcome.values()))

    if isinstance(outcome, OutcomeType):
        return outcome.astuple()

    if isinstance(outcome, np.ndarray):
        return tuple(outcome.tolist())

    if isinstance(outcome, Iterable):
        return tuple(outcome)

    raise ValueError(f"Unknown type for outcome {type(outcome)}")


def outcome_as(outcome: "Outcome", astype: Type, issue_names: List[str] = None):
    """Converts the outcome to tuple, dict or any `OutcomeType`.

    Args:
         outcome: The outcome to adjust type
         astype: The type to return. None returns the outcome as it is
         issue_names: Only needed if `astype` is not tuple

    """
    if astype is None:
        return outcome
    if issubclass(astype, tuple):
        return outcome_as_tuple(outcome)
    if issubclass(astype, dict):
        return outcome_as_dict(outcome, issue_names)
    return astype(**outcome_as_dict(outcome, issue_names))


def outcome_for(outcome: "Outcome", ami: "Mechanism") -> Optional["Outcome"]:
    """Converts the outcome the type specified by the mechanism

    Args:
         outcome: The outcome to adjust type
         ami: The Agent Mechanism Interface

    """
    astype = ami.outcome_type
    issue_names = None if issubclass(astype, tuple) else [_.name for _ in ami.issues]
    return outcome_as(outcome, astype, issue_names)



import logging
import math
import random
from pprint import pformat

import pandas as pd
from typing import Any, Dict

from .baseline import DummyElicitor, FullKnowledgeElicitor
from .expectors import BalancedExpector, MaxExpector, MinExpector
from .pandora import FullElicitor, RandomElicitor
from .user import User
from .voi import (
    Answer,
    BaseElicitor,
    EStrategy,
    MeanExpector,
    Optional,
    Query,
    RangeConstraint,
    VOIElicitor,
    VOIFastElicitor,
    VOINoUncertaintyElicitor,
    VOIOptimalElicitor,
    np,
    time,
)
from ..genius import GeniusNegotiator
from ..helpers import create_loggers, instantiate
from ..inout import load_genius_domain_from_folder
from ..mechanisms import Mechanism
from ..modeling import UncertainOpponentModel
from ..negotiators import AspirationMixin
from ..outcomes import Outcome
from ..sao import (
    AspirationNegotiator,
    LimitedOutcomesAcceptor,
    OnlyBestNegotiator,
    RandomNegotiator,
    SAOState,
    ToughNegotiator,
    SAOMechanism,
    LimitedOutcomesNegotiator,
)
from ..utilities import (
    IPUtilityFunction,
    MappingUtilityFunction,
    UtilityDistribution,
    UtilityFunction,
    UtilityValue,
)

__all__ = ["SAOElicitingMechanism"]


def uniform():
    loc = random.random()
    scale = random.random() * (1.0 - loc)
    return UtilityDistribution(dtype="uniform", loc=loc, scale=scale)


def current_aspiration(
    elicitor: "AspirationMixin", outcome: "Outcome", negotiation: "Mechanism"
) -> "UtilityValue":
    return elicitor.aspiration(negotiation.relative_time)


def create_negotiator(
    negotiator_type, ufun, can_propose, outcomes, toughness, **kwargs
):
    if negotiator_type == "limited_outcomes":
        if can_propose:
            negotiator = LimitedOutcomesNegotiator(
                acceptable_outcomes=outcomes,
                acceptance_probabilities=list(ufun.mapping.values()),
                **kwargs,
            )
        else:
            negotiator = LimitedOutcomesAcceptor(
                acceptable_outcomes=outcomes,
                acceptance_probabilities=list(ufun.mapping.values()),
                **kwargs,
            )
    elif negotiator_type == "random":
        negotiator = RandomNegotiator(can_propose=can_propose,)
    elif negotiator_type == "tough":
        negotiator = ToughNegotiator(can_propose=can_propose)
    elif negotiator_type in ("only_best", "best_only", "best"):
        negotiator = OnlyBestNegotiator(
            min_utility=None,
            top_fraction=1.0 - toughness,
            best_first=False,
            can_propose=can_propose,
        )
    elif negotiator_type.startswith("aspiration"):
        asp_kind = negotiator_type[len("aspiration") :]
        if asp_kind.startswith("_"):
            asp_kind = asp_kind[1:]
        try:
            asp_kind = float(asp_kind)
        except:
            pass
        if asp_kind == "":
            if toughness < 0.5:
                toughness *= 2
                toughness = 9.0 * toughness + 1.0
            elif toughness == 0.5:
                toughness = 1.0
            else:
                toughness = 2 * (toughness - 0.5)
                toughness = 1 - 0.9 * toughness
            asp_kind = toughness
        negotiator = AspirationNegotiator(
            aspiration_type=asp_kind, can_propose=can_propose, **kwargs,
        )
    elif negotiator_type.startswith("genius"):
        class_name = negotiator_type[len("genius") :]
        if class_name.startswith("_"):
            class_name = class_name[1:]
        if class_name == "auto" or len(class_name) < 1:
            negotiator = GeniusNegotiator.random_negotiator(
                keep_issue_names=False, keep_value_names=False, can_propose=can_propose
            )
        else:
            negotiator = GeniusNegotiator(
                java_class_name=class_name,
                keep_value_names=False,
                keep_issue_names=False,
                can_propose=can_propose,
            )
        negotiator.utility_function = ufun
    else:
        raise ValueError(f"Unknown opponents type {negotiator_type}")
    return negotiator


def _beg(x):
    if isinstance(x, float):
        return x
    else:
        return x.loc


def _scale(x):
    if isinstance(x, float):
        return 0.0
    else:
        return x.scale


def _end(x):
    if isinstance(x, float):
        return x
    else:
        return x.loc + x.scale


class SAOElicitingMechanism(SAOMechanism):
    @classmethod
    def generate_config(
        cls,
        cost,
        n_outcomes: int = None,
        rand_ufuns=True,
        conflict: float = None,
        conflict_delta: float = None,
        winwin=None,  # only if rand_ufuns is false
        genius_folder: str = None,
        n_steps=None,
        time_limit=None,
        own_utility_uncertainty=0.5,
        own_uncertainty_variablility=0.0,
        own_reserved_value=0.0,
        own_base_agent="aspiration",
        opponent_model_uncertainty=0.5,
        opponent_model_adaptive=False,
        opponent_proposes=True,
        opponent_type="best_only",
        opponent_toughness=0.9,
        opponent_reserved_value=0.0,
    ) -> Dict[str, Any]:
        config = {}
        if n_steps is None and time_limit is None and "aspiration" in opponent_type:
            raise ValueError(
                "Cannot use aspiration negotiators when no step limit or time limit is given"
            )
        if n_outcomes is None and genius_folder is None:
            raise ValueError(
                "Must specify a folder to run from or a number of outcomes"
            )
        if genius_folder is not None:
            domain, agent_info, issues = load_genius_domain_from_folder(
                folder_name=genius_folder,
                force_single_issue=True,
                keep_issue_names=False,
                keep_value_names=False,
                ignore_discount=True,
                ignore_reserved=opponent_reserved_value is not None,
            )
            n_outcomes = domain.ami.n_outcomes
            outcomes = domain.outcomes
            elicitor_indx = 0 + int(random.random() <= 0.5)
            opponent_indx = 1 - elicitor_indx
            ufun = agent_info[elicitor_indx]["ufun"]
            ufun.reserved_value = own_reserved_value
            opp_utility = agent_info[opponent_indx]["ufun"]
            opp_utility.reserved_value = opponent_reserved_value
        else:
            outcomes = [(_,) for _ in range(n_outcomes)]
            if rand_ufuns:
                ufun, opp_utility = UtilityFunction.generate_random_bilateral(
                    outcomes=outcomes
                )
            else:
                ufun, opp_utility = UtilityFunction.generate_bilateral(
                    outcomes=outcomes,
                    conflict_level=opponent_toughness,
                    conflict_delta=conflict_delta,
                    win_win=winwin,
                )
            ufun.reserved_value = own_reserved_value
            domain = SAOMechanism(
                outcomes=outcomes,
                n_steps=n_steps,
                time_limit=time_limit,
                max_n_agents=2,
                dynamic_entry=False,
                outcome_type=tuple,
                cache_outcomes=True,
            )

        true_utilities = list(ufun.mapping.values())
        priors = IPUtilityFunction.from_ufun(
            ufun,
            uncertainty=own_utility_uncertainty,
            variability=own_uncertainty_variablility,
        )

        outcomes = domain.ami.outcomes

        opponent = create_negotiator(
            negotiator_type=opponent_type,
            can_propose=opponent_proposes,
            ufun=opp_utility,
            outcomes=outcomes,
            toughness=opponent_toughness,
        )
        opponent_model = UncertainOpponentModel(
            outcomes=outcomes,
            uncertainty=opponent_model_uncertainty,
            opponents=opponent,
            adaptive=opponent_model_adaptive,
        )
        config["n_steps"], config["time_limit"] = n_steps, time_limit
        config["priors"] = priors
        config["true_utilities"] = true_utilities
        config["elicitor_reserved_value"] = own_reserved_value
        config["cost"] = cost
        config["opp_utility"] = opp_utility
        config["opponent_model"] = opponent_model
        config["opponent"] = opponent
        config["base_agent"] = own_base_agent
        return config

    def __init__(
        self,
        priors,
        true_utilities,
        elicitor_reserved_value,
        cost,
        opp_utility,
        opponent,
        n_steps,
        time_limit,
        base_agent,
        opponent_model,
        elicitation_strategy="pingpong",
        toughness=0.95,
        elicitor_type="balanced",
        history_file_name: str = None,
        screen_log: bool = False,
        dynamic_queries=True,
        each_outcome_once=False,
        rational_answer_probs=True,
        update_related_queries=True,
        resolution=0.1,
        cost_assuming_titration=False,
        name: Optional[str] = None,
    ):
        self.elicitation_state = {}
        initial_priors = priors
        self.xw_real = priors

        outcomes = list(initial_priors.distributions.keys())

        self.U = true_utilities

        super().__init__(
            issues=None,
            outcomes=outcomes,
            n_steps=n_steps,
            time_limit=time_limit,
            max_n_agents=2,
            dynamic_entry=False,
            name=name,
            outcome_type=tuple,
            enable_callbacks=True,
        )
        if elicitor_reserved_value is None:
            elicitor_reserved_value = 0.0
        self.logger = create_loggers(
            file_name=history_file_name,
            screen_level=logging.DEBUG if screen_log else logging.ERROR,
        )
        user = User(
            ufun=MappingUtilityFunction(
                dict(zip(self.outcomes, self.U)), reserved_value=elicitor_reserved_value
            ),
            cost=cost,
            ami=self.ami,
        )
        if resolution is None:
            resolution = max(elicitor_reserved_value / 4, 0.025)
        if "voi" in elicitor_type and "optimal" in elicitor_type:
            strategy = None
        else:
            strategy = EStrategy(strategy=elicitation_strategy, resolution=resolution)
            strategy.on_enter(ami=self.ami, ufun=initial_priors)

        def create_elicitor(type_, strategy=strategy, opponent_model=opponent_model):
            base_negotiator = create_negotiator(
                negotiator_type=base_agent,
                ufun=None,
                can_propose=True,
                outcomes=outcomes,
                toughness=toughness,
            )
            if type_ == "full":
                return FullElicitor(
                    strategy=strategy, user=user, base_negotiator=base_negotiator
                )

            if type_ == "dummy":
                return DummyElicitor(
                    strategy=strategy, user=user, base_negotiator=base_negotiator
                )

            if type_ == "full_knowledge":
                return FullKnowledgeElicitor(
                    strategy=strategy, user=user, base_negotiator=base_negotiator
                )

            if type_ == "random_deep":
                return RandomElicitor(
                    strategy=strategy,
                    deep_elicitation=True,
                    user=user,
                    base_negotiator=base_negotiator,
                )

            if type_ in ("random_shallow", "random"):
                return RandomElicitor(
                    strategy=strategy,
                    deep_elicitation=False,
                    user=user,
                    base_negotiator=base_negotiator,
                )
            if type_ in (
                "pessimistic",
                "optimistic",
                "balanced",
                "pandora",
                "fast",
                "mean",
            ):
                type_ = type_.title() + "Elicitor"
                return instantiate(
                    f"negmas.elicitation.{type_}",
                    strategy=strategy,
                    user=user,
                    base_negotiator=base_negotiator,
                    opponent_model_factory=lambda x: opponent_model,
                    single_elicitation_per_round=False,
                    assume_uniform=True,
                    user_model_in_index=True,
                    precalculated_index=False,
                )
            if "voi" in type_:
                expector_factory = MeanExpector
                if "balanced" in type_:
                    expector_factory = BalancedExpector
                elif "optimistic" in type_ or "max" in type_:
                    expector_factory = MaxExpector
                elif "pessimistic" in type_ or "min" in type_:
                    expector_factory = MinExpector

                if "fast" in type_:
                    factory = VOIFastElicitor
                elif "optimal" in type_:
                    prune = "prune" in type_ or "fast" in type_
                    if "no" in type_:
                        no_prune = not prune
                    return VOIOptimalElicitor(
                        user=user,
                        resolution=resolution,
                        opponent_model_factory=lambda x: opponent_model,
                        single_elicitation_per_round=False,
                        base_negotiator=base_negotiator,
                        each_outcome_once=each_outcome_once,
                        expector_factory=expector_factory,
                        update_related_queries=update_related_queries,
                        prune=prune,
                    )
                elif "no_uncertainty" in type_ or "full_knowledge" in type_:
                    factory = VOINoUncertaintyElicitor
                else:
                    factory = VOIElicitor

                if not dynamic_queries and "optimal" not in type_:
                    queries = []
                    for outcome in self.outcomes:
                        u = initial_priors(outcome)
                        scale = _scale(u)
                        if scale < resolution:
                            continue
                        bb, ee = _beg(u), _end(u)
                        n_q = int((ee - bb) / resolution)
                        limits = np.linspace(bb, ee, n_q, endpoint=False)[1:]
                        for i, limit in enumerate(limits):
                            if cost_assuming_titration:
                                qcost = cost * min(i, len(limits) - i - 1)
                            else:
                                qcost = cost
                            answers = [
                                Answer(
                                    outcomes=[outcome],
                                    constraint=RangeConstraint(rng=(0.0, limit)),
                                    name="yes",
                                ),
                                Answer(
                                    outcomes=[outcome],
                                    constraint=RangeConstraint(rng=(limit, 1.0)),
                                    name="no",
                                ),
                            ]
                            probs = (
                                [limit, 1.0 - limit]
                                if rational_answer_probs
                                else [0.5, 0.5]
                            )
                            query = Query(
                                answers=answers,
                                cost=qcost,
                                probs=probs,
                                name=f"{outcome}<{limit}",
                            )
                            queries.append((outcome, query, qcost))
                else:
                    queries = None
                return factory(
                    strategy=strategy if dynamic_queries else None,
                    user=user,
                    opponent_model_factory=lambda x: opponent_model,
                    single_elicitation_per_round=False,
                    dynamic_query_set=dynamic_queries,
                    queries=queries,
                    base_negotiator=base_negotiator,
                    each_outcome_once=each_outcome_once,
                    expector_factory=expector_factory,
                    update_related_queries=update_related_queries,
                )

        elicitor = create_elicitor(elicitor_type)

        if isinstance(opponent, GeniusNegotiator):
            if n_steps is not None and time_limit is not None:
                self.ami.n_steps = None

        self.add(opponent, ufun=opp_utility)
        self.add(elicitor, ufun=initial_priors)
        if len(self.negotiators) != 2:
            raise ValueError(
                f"I could not add the two negotiators {elicitor.__class__.__name__}, {opponent.__class__.__name__}"
            )
        self.total_time = 0.0

    def loginfo(self, s: str) -> None:
        """logs ami-level information

        Args:
            s (str): The string to log

        """
        self.logger.info(s.strip())

    def logdebug(self, s) -> None:
        """logs debug-level information

        Args:
            s (str): The string to log

        """
        self.logger.debug(s.strip())

    def logwarning(self, s) -> None:
        """logs warning-level information

        Args:
            s (str): The string to log

        """
        self.logger.warning(s.strip())

    def logerror(self, s) -> None:
        """logs error-level information

        Args:
            s (str): The string to log

        """
        self.logger.error(s.strip())

    def step(self) -> SAOState:
        start = time.perf_counter()
        _ = super().step()
        self.total_time += time.perf_counter() - start
        self.loginfo(
            f"[{self._step}] {self._current_proposer} offered {self._current_offer}"
        )
        return _

    def on_negotiation_start(self):
        if not super().on_negotiation_start():
            return False
        self.elicitation_state = {}
        self.elicitation_state["steps"] = None
        self.elicitation_state["relative_time"] = None
        self.elicitation_state["broken"] = False
        self.elicitation_state["timedout"] = False
        self.elicitation_state["agreement"] = None
        self.elicitation_state["agreed"] = False
        self.elicitation_state["utils"] = [
            0.0 for a in self.negotiators
        ]  # not even the reserved value
        self.elicitation_state["welfare"] = sum(self.elicitation_state["utils"])
        self.elicitation_state["elicitor"] = self.negotiators[
            1
        ].__class__.__name__.replace("Elicitor", "")
        self.elicitation_state["opponents"] = self.negotiators[
            0
        ].__class__.__name__.replace("Aget", "")
        self.elicitation_state["elicitor_utility"] = self.elicitation_state["utils"][1]
        self.elicitation_state["opponent_utility"] = self.elicitation_state["utils"][0]
        self.elicitation_state["opponent_params"] = str(self.negotiators[0])
        self.elicitation_state["elicitor_params"] = str(self.negotiators[1])
        self.elicitation_state["elicitation_cost"] = None
        self.elicitation_state["total_time"] = None
        self.elicitation_state["pareto"] = None
        self.elicitation_state["pareto_distance"] = None
        self.elicitation_state["_elicitation_time"] = None
        self.elicitation_state["real_asking_time"] = None
        self.elicitation_state["n_queries"] = 0
        return True

    def plot(self, consider_costs=False):
        try:
            import matplotlib.pyplot as plt
            import matplotlib.gridspec as gridspec

            if len(self.negotiators) > 2:
                print("Cannot visualize negotiations with more than 2 negotiators")
            else:
                # has_front = int(len(self.outcomes[0]) <2)
                has_front = 1
                n_agents = len(self.negotiators)
                history = pd.DataFrame(data=[_[1] for _ in self.history])
                history["time"] = [_[0].time for _ in self.history]
                history["relative_time"] = [_[0].relative_time for _ in self.history]
                history["step"] = [_[0].step for _ in self.history]
                history = history.loc[~history.offer.isnull(), :]
                ufuns = self._get_ufuns(consider_costs=consider_costs)
                elicitor_dist = self.negotiators[1].utility_function
                outcomes = self.outcomes

                utils = [tuple(f(o) for f in ufuns) for o in outcomes]
                agent_names = [
                    a.__class__.__name__ + ":" + a.name for a in self.negotiators
                ]
                history["offer_index"] = [outcomes.index(_) for _ in history.offer]
                frontier, frontier_outcome = self.pareto_frontier(sort_by_welfare=True)
                frontier_outcome_indices = [outcomes.index(_) for _ in frontier_outcome]
                fig_util, fig_outcome = plt.figure(), plt.figure()
                gs_util = gridspec.GridSpec(n_agents, has_front + 1)
                gs_outcome = gridspec.GridSpec(n_agents, has_front + 1)
                axs_util, axs_outcome = [], []

                agent_names_for_legends = [
                    agent_names[a]
                    .split(":")[0]
                    .replace("Negotiator", "")
                    .replace("Elicitor", "")
                    for a in range(n_agents)
                ]
                if agent_names_for_legends[0] == agent_names_for_legends[1]:
                    agent_names_for_legends = [
                        agent_names[a]
                        .split(":")[0]
                        .replace("Negotiator", "")
                        .replace("Elicitor", "")
                        + agent_names[a].split(":")[1]
                        for a in range(n_agents)
                    ]

                for a in range(n_agents):
                    if a == 0:
                        axs_util.append(fig_util.add_subplot(gs_util[a, has_front]))
                        axs_outcome.append(
                            fig_outcome.add_subplot(gs_outcome[a, has_front])
                        )
                    else:
                        axs_util.append(
                            fig_util.add_subplot(
                                gs_util[a, has_front], sharex=axs_util[0]
                            )
                        )
                        axs_outcome.append(
                            fig_outcome.add_subplot(
                                gs_outcome[a, has_front], sharex=axs_outcome[0]
                            )
                        )
                    axs_util[-1].set_ylabel(agent_names_for_legends[a])
                    axs_outcome[-1].set_ylabel(agent_names_for_legends[a])
                for a, (au, ao) in enumerate(zip(axs_util, axs_outcome)):
                    h = history.loc[
                        history.offerer == agent_names[a],
                        ["relative_time", "offer_index", "offer"],
                    ]
                    h["utility"] = h.offer.apply(ufuns[a])
                    ao.plot(h.relative_time, h.offer_index)
                    au.plot(h.relative_time, h.utility)
                    # if a == 1:
                    h["dist"] = h.offer.apply(elicitor_dist)
                    h["beg"] = h.dist.apply(_beg)
                    h["end"] = h.dist.apply(_end)
                    h["p_acceptance"] = h.offer.apply(
                        self.negotiators[1].opponent_model.probability_of_acceptance
                    )
                    au.plot(h.relative_time, h.end, color="r")
                    au.plot(h.relative_time, h.beg, color="r")
                    au.plot(h.relative_time, h.p_acceptance, color="g")
                    au.set_ylim(-0.1, 1.1)

                if has_front:
                    axu = fig_util.add_subplot(gs_util[:, 0])
                    axu.plot([0, 1], [0, 1], "g--")
                    axu.scatter(
                        [_[0] for _ in utils],
                        [_[1] for _ in utils],
                        label="outcomes",
                        color="yellow",
                        marker="s",
                        s=20,
                    )
                    axo = fig_outcome.add_subplot(gs_outcome[:, 0])
                    clrs = ("blue", "green")
                    for a in range(n_agents):
                        h = history.loc[
                            history.offerer == agent_names[a],
                            ["relative_time", "offer_index", "offer"],
                        ]
                        h["u0"] = h.offer.apply(ufuns[0])
                        h["u1"] = h.offer.apply(ufuns[1])

                        axu.scatter(
                            h.u0,
                            h.u1,
                            color=clrs[a],
                            label=f"{agent_names_for_legends[a]}",
                        )
                    steps = sorted(history.step.unique().tolist())
                    aoffers = [[], []]
                    for step in steps[::2]:
                        offrs = []
                        for a in range(n_agents):
                            a_offer = history.loc[
                                (history.offerer == agent_names[a])
                                & ((history.step == step) | (history.step == step + 1)),
                                "offer_index",
                            ]
                            if len(a_offer) > 0:
                                offrs.append(a_offer.values[-1])
                        if len(offrs) == 2:
                            aoffers[0].append(offrs[0])
                            aoffers[1].append(offrs[1])
                    axo.scatter(aoffers[0], aoffers[1], color=clrs[0], label=f"offers")

                    if self.state.agreement is not None:
                        axu.scatter(
                            [ufuns[0](self.state.agreement)],
                            [ufuns[1](self.state.agreement)],
                            color="black",
                            marker="*",
                            s=120,
                            label="SCMLAgreement",
                        )
                        axo.scatter(
                            [outcomes.index(self.state.agreement)],
                            [outcomes.index(self.state.agreement)],
                            color="black",
                            marker="*",
                            s=120,
                            label="SCMLAgreement",
                        )
                    f1, f2 = [_[0] for _ in frontier], [_[1] for _ in frontier]
                    axu.scatter(f1, f2, label="frontier", color="red", marker="x")
                    axo.scatter(
                        frontier_outcome_indices,
                        frontier_outcome_indices,
                        color="red",
                        marker="x",
                        label="frontier",
                    )
                    axu.legend()
                    axo.legend()
                    axo.set_xlabel(agent_names_for_legends[0])
                    axo.set_ylabel(agent_names_for_legends[1])

                    axu.set_xlabel(agent_names_for_legends[0] + " utility")
                    axu.set_ylabel(agent_names_for_legends[1] + " utility")
                    if self.agreement is not None:
                        pareto_distance = 1e9
                        cu = (ufuns[0](self.agreement), ufuns[1](self.agreement))
                        for pu in frontier:
                            dist = math.sqrt(
                                (pu[0] - cu[0]) ** 2 + (pu[1] - cu[1]) ** 2
                            )
                            if dist < pareto_distance:
                                pareto_distance = dist
                        axu.text(
                            0,
                            0.95,
                            f"Pareto-distance={pareto_distance:5.2}",
                            verticalalignment="top",
                            transform=axu.transAxes,
                        )

                fig_util.show()
                fig_outcome.show()
        except:
            pass

    def on_negotiation_end(self):
        super().on_negotiation_end()
        self.elicitation_state = {}
        self.elicitation_state["steps"] = self._step + 1
        self.elicitation_state["relative_time"] = self.relative_time
        self.elicitation_state["broken"] = self.state.broken
        self.elicitation_state["timedout"] = (
            not self.state.broken and self.state.agreement is None
        )
        self.elicitation_state["agreement"] = self.state.agreement
        self.elicitation_state["agreed"] = (
            self.state.agreement is not None and not self.state.broken
        )

        if self.elicitation_state["agreed"]:
            self.elicitation_state["utils"] = [
                a.user_ufun(self.state.agreement)
                if isinstance(a, BaseElicitor)
                else a.ufun(self.state.agreement)
                for a in self.negotiators
            ]
        else:
            self.elicitation_state["utils"] = [
                a.reserved_value if a.reserved_value is not None else 0.0
                for a in self.negotiators
            ]
        self.elicitation_state["welfare"] = sum(self.elicitation_state["utils"])
        self.elicitation_state["elicitor"] = self.negotiators[
            1
        ].__class__.__name__.replace("Elicitor", "")
        self.elicitation_state["opponents"] = self.negotiators[
            0
        ].__class__.__name__.replace("Aget", "")
        self.elicitation_state["elicitor_utility"] = self.elicitation_state["utils"][1]
        self.elicitation_state["opponent_utility"] = self.elicitation_state["utils"][0]
        self.elicitation_state["opponent_params"] = str(self.negotiators[0])
        self.elicitation_state["elicitor_params"] = str(self.negotiators[1])
        self.elicitation_state["elicitation_cost"] = self.negotiators[
            1
        ].elicitation_cost
        self.elicitation_state["total_time"] = self.total_time
        self.elicitation_state["_elicitation_time"] = self.negotiators[
            1
        ].elicitation_time
        self.elicitation_state["asking_time"] = self.negotiators[1].asking_time
        self.elicitation_state["pareto"], pareto_outcomes = self.pareto_frontier()
        if self.elicitation_state["agreed"]:
            if self.state.agreement in pareto_outcomes:
                min_dist = 0.0
            else:
                min_dist = 1e12
                for p in self.elicitation_state["pareto"]:
                    dist = 0.0
                    for par, real in zip(p, self.elicitation_state["utils"]):
                        dist += (par - real) ** 2
                    dist = math.sqrt(dist)
                    if dist < min_dist:
                        min_dist = dist
            self.elicitation_state["pareto_distance"] = (
                min_dist if min_dist < 1e12 else None
            )
        else:
            self.elicitation_state["pareto_distance"] = None
        try:
            self.elicitation_state["queries"] = [
                str(_) for _ in self.negotiators[1].user.elicited_queries()
            ]
        except:
            self.elicitation_state["queries"] = None
        try:
            self.elicitation_state["n_queries"] = len(
                self.negotiators[1].user.elicited_queries()
            )
        except:
            self.elicitation_state["n_queries"] = None
        if hasattr(self.negotiators[1], "total_voi"):
            self.elicitation_state["total_voi"] = self.negotiators[1].total_voi
        else:
            self.elicitation_state["total_voi"] = None




from dataclasses import dataclass

import numpy as np
from typing import List, Optional

from ..outcomes import Outcome
from ..utilities import UtilityFunction, MappingUtilityFunction
from .queries import Query, QResponse, Constraint, CostEvaluator

np.seterr(all="raise")  # setting numpy to raise exceptions in case of errors

__all__ = ["User", "ElicitationRecord"]


@dataclass
class ElicitationRecord:
    cost: float
    query: Query
    answer_index: int
    step: Optional[int] = None

    def __str__(self):
        if self.step is None:
            return f"{self.query}: {self.query.answers[self.answer_index]} @{self.cost}"
        return f"[{self.step}] {self.query}: {self.query.answers[self.answer_index]} @{self.cost}"

    __repr__ = __str__


class User:
    """Abstract base class for all representations of users used for elicitation

    Args:
        ufun: The real utility function of the user.
        cost: A cost to be added for every question asked to the user.
        ami: [Optional] The `AgentMechanismInterface` representing *the*
             negotiation session engaged in by this user using this `ufun`.

    """

    def __init__(
        self, ufun: Optional[UtilityFunction] = None, cost: float = 0.0, ami=None
    ):
        super().__init__()
        self.utility_function = ufun
        self.cost = cost
        self.total_cost = 0.0
        self._elicited_queries: List[ElicitationRecord] = []
        self.ami = ami

    def set(self, ufun: Optional[UtilityFunction] = None, cost: float = None):
        """
        Sets the ufun and/or cost for this user        <`0:desc`>

        Args:
            ufun: The ufun if given
            cost: The cost if given
        """

        if ufun is not None:
            self.utility_function = ufun
        if cost is not None:
            self.cost = cost

    @property
    def ufun(self) -> UtilityFunction:
        """Gets a `UtilityFunction` representing the real utility_function of the user"""
        return (
            self.utility_function
            if self.utility_function is not None
            else MappingUtilityFunction(lambda x: None)
        )

    def ask(self, q: Optional[Query]) -> QResponse:
        """Query the user and get a response."""
        if q is None:
            return QResponse(answer=None, indx=-1, cost=0.0)
        self.total_cost += self.cost + q.cost
        for i, reply in enumerate(q.answers):
            if reply.constraint.is_satisfied(self.ufun, reply.outcomes):
                self.total_cost += reply.cost
                self._elicited_queries.append(
                    ElicitationRecord(
                        query=q,
                        cost=self.cost + q.cost + reply.cost,
                        answer_index=i,
                        step=self.ami.state.step if self.ami else None,
                    )
                )
                return QResponse(
                    answer=reply, indx=i, cost=CostEvaluator(self.cost)(q, reply)
                )
        print(f"No response for {q} (ufun={self.ufun})")
        return QResponse(answer=None, indx=-1, cost=q.cost)

    def cost_of_asking(
        self, q: Optional[Query] = None, answer_id: int = -1, estimate_answer_cost=True
    ) -> float:
        """
        Returns the cost of asking the given `Quers`.

        Args:
            q: The query
            answer_id: If >= 0, the answer expected. Used to add the specific
                      cost of the answer if `estimate_answer_cost` is `True`
            estimate_answer_cost: If `True` and `answer_id` >= 0, the specific
                                  cost of getting this answer is added.
        """
        if q is None:
            return self.cost
        cost = self.cost + q.cost
        if not estimate_answer_cost:
            return cost
        if answer_id <= 0:
            return cost + q.answers[answer_id].cost
        return cost + sum(a.cost for a in q.answers) / len(q.answers)

    def is_satisfied(self, constraint: Constraint, outcomes=List[Outcome]) -> bool:
        """Checks if the given consgtraint is satisfied for the user
        utility fun and the given outocmes.

        Args:
            constraint: The `Constraint`
            outcome: A list of `Outcome`s to be passed to the constraint along
                     with the user's ufun.
        """
        return constraint.is_satisfied(self.ufun, outcomes=outcomes)

    def elicited_queries(self) -> List[ElicitationRecord]:
        """Returns a list of elicited queries.

        For each elicited query, the following dataclass is returned:
        ElicitationRecord(query, cost, answer_index, step)
        """
        return self._elicited_queries



import copy
import operator
import pprint
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Callable, Iterable, List, Optional, Sequence, Tuple, Union

from .common import _loc, _upper
from ..common import AgentMechanismInterface
from ..outcomes import Outcome
from ..utilities import (
    UtilityDistribution,
    UtilityFunction,
    UtilityValue,
)

__all__ = [
    "Constraint",
    "MarginalNeutralConstraint",
    "RankConstraint",
    "ComparisonConstraint",
    "RangeConstraint",
    "Answer",
    "Query",
    "QResponse",
    "next_query",
    "possible_queries",
    "CostEvaluator",
]


class Constraint(ABC):
    """Some constraint on allowable utility values for given outcomes."""

    def __init__(
        self,
        full_range: Union[Sequence[Tuple[float, float]], Tuple[float, float]] = (
            0.0,
            1.0,
        ),
        outcomes: List[Outcome] = None,
    ):
        super().__init__()
        self.outcomes = outcomes
        self.index = None
        if outcomes is not None:
            self.index = dict(zip(outcomes, range(len(outcomes))))
            if not isinstance(full_range, tuple):
                full_range = [full_range] * len(outcomes)
        self.full_range = full_range

    @abstractmethod
    def is_satisfied(
        self, ufun: UtilityFunction, outcomes: Optional[Iterable[Outcome]] = None
    ) -> bool:
        ...

    def __str__(self):
        return pprint.pformat(self.__dict__)

    def __repr__(self):
        return self.__dict__.__repr__()

    @abstractmethod
    def marginals(
        self, outcomes: Iterable[Outcome] = None
    ) -> List[UtilityDistribution]:
        ...

    @abstractmethod
    def marginal(self, outcome: "Outcome") -> UtilityDistribution:
        ...


class MarginalNeutralConstraint(Constraint):
    """Constraints that do not affect the marginals of any outcomes. These constraints may only affect the joint
    distribution."""

    def marginals(
        self, outcomes: Iterable[Outcome] = None
    ) -> List[UtilityDistribution]:
        if outcomes is None:
            outcomes = self.outcomes
        # this works only for real-valued outcomes.
        return [
            UtilityDistribution(
                dtype="uniform",
                loc=self.full_range[_][0],
                scale=self.full_range[_][1] - self.full_range[_][0],
            )
            for _ in range(len(outcomes))
        ]

    def marginal(self, outcome: "Outcome") -> UtilityDistribution:
        # this works only for real-valued outcomes.
        if self.outcomes is None:
            return UtilityDistribution(
                dtype="uniform",
                loc=self.full_range[0],
                scale=self.full_range[1] - self.full_range[0],
            )
        indx = self.index[outcome]
        return UtilityDistribution(
            dtype="uniform",
            loc=self.full_range[indx][0],
            scale=self.full_range[indx][1] - self.full_range[indx][0],
        )


class RankConstraint(MarginalNeutralConstraint):
    """Constraints the utilities of given outcomes to be in ascending order
    """

    def __init__(
        self,
        rankings: List[int],
        full_range: Union[Sequence[Tuple[float, float]], Tuple[float, float]] = (
            0.0,
            1.0,
        ),
        outcomes: List[Outcome] = None,
    ):
        super().__init__(full_range=full_range, outcomes=outcomes)
        self.rankings = rankings

    def is_satisfied(
        self, ufun: UtilityFunction, outcomes: Optional[Iterable[Outcome]] = None
    ) -> bool:
        if outcomes is None:
            outcomes = self.outcomes
        if outcomes is None:
            raise ValueError("No outcomes are  given in construction or to the call")
        u = [(ufun(o), i) for i, o in enumerate(outcomes)]
        ranking = sorted(u, key=lambda x: x[0])
        return ranking == self.rankings


class ComparisonConstraint(MarginalNeutralConstraint):
    """Constraints the utility of given two outcomes (must be exactly two) to satisfy the given operation (e.g. >, <)"""

    def __init__(
        self,
        op: Union[str, Callable[[UtilityValue, UtilityValue], bool]],
        full_range: Union[Sequence[Tuple[float, float]], Tuple[float, float]] = (
            0.0,
            1.0,
        ),
        outcomes: List[Outcome] = None,
    ):
        super().__init__(full_range=full_range, outcomes=outcomes)
        if outcomes is not None and len(outcomes) != 2:
            raise ValueError(
                f"{len(outcomes)} outcomes were given to {self.__class__.__name__}"
            )
        self.op_name = op
        if isinstance(op, str):
            if op in ("less", "l", "<"):
                op = operator.lt
            elif op in ("greater", "g", ">"):
                op = operator.gt
            elif op in ("equal", "=", "=="):
                op = operator.eq
            elif op in ("le", "<="):
                op = operator.le
            elif op in ("ge", ">="):
                op = operator.ge
            else:
                raise ValueError(f"Unknown operation {op}")
        self.op = op

    def is_satisfied(
        self, ufun: UtilityFunction, outcomes: Optional[Iterable[Outcome]] = None
    ) -> bool:
        if outcomes is None:
            outcomes = self.outcomes
        if outcomes is None:
            raise ValueError("No outcomes are  given in construction or to the call")
        if len(outcomes) != 2:
            raise ValueError(
                f"{len(outcomes)} outcomes were given to {self.__class__.__name__}"
            )
        u = [(ufun(o), i) for i, o in enumerate(outcomes)]
        return self.op(u[0], u[1])

    def __str__(self):
        return f"{self.outcomes[0]} {self.op_name} {self.outcomes[0]}"

    __repr__ = __str__


class RangeConstraint(Constraint):
    """Constraints the utility of each of the given outcomes to lie within the given range"""

    def __init__(
        self,
        rng: Tuple = (None, None),
        full_range: Union[Sequence[Tuple[float, float]], Tuple[float, float]] = (
            0.0,
            1.0,
        ),
        outcomes: List[Outcome] = None,
        eps=1e-5,
    ):
        super().__init__(full_range=full_range, outcomes=outcomes)

        if outcomes is not None:
            self.index = dict(zip(outcomes, range(len(outcomes))))
            if not isinstance(rng, tuple):
                rng = [rng] * len(outcomes)
        self.range = rng
        self.eps = eps
        if outcomes is None:
            self.effective_range = (
                rng[0] if rng[0] is not None else self.full_range[0],
                rng[1] if rng[1] is not None else self.full_range[1],
            )
        else:
            self.effective_range = [
                (r[0] if r[0] is not None else f[0], r[1] if r[1] is not None else f[1])
                for r, f in zip(self.range, self.full_range)
            ]

    def is_satisfied(
        self, ufun: UtilityFunction, outcomes: Optional[Iterable[Outcome]] = None
    ) -> bool:
        if outcomes is None:
            outcomes = self.outcomes
        if outcomes is None:
            raise ValueError("No outcomes are  given in construction or to the call")
        us = [ufun(o) for o in outcomes]
        mn, mx = self.range
        if mn is not None:
            for u in us:
                if u < mn - self.eps:
                    return False
        if mx is not None:
            for u in us:
                if u > mx + self.eps:
                    return False
        return True

    def marginals(
        self, outcomes: Iterable[Outcome] = None
    ) -> List[UtilityDistribution]:
        if outcomes is None:
            outcomes = self.outcomes
        # this works only for real-valued outcomes.
        return [
            UtilityDistribution(
                dtype="uniform",
                loc=self.effective_range[_][0],
                scale=self.effective_range[_][1] - self.effective_range[_][0],
            )
            for _ in range(len(outcomes))
        ]

    def marginal(self, outcome: "Outcome") -> UtilityDistribution:
        # this works only for real-valued outcomes.
        if self.outcomes is None:
            return UtilityDistribution(
                dtype="uniform",
                loc=self.effective_range[0],
                scale=self.effective_range[1] - self.effective_range[0],
            )
        indx = self.index[outcome]
        return UtilityDistribution(
            dtype="uniform",
            loc=self.effective_range[indx][0],
            scale=self.effective_range[indx][1] - self.effective_range[indx][0],
        )

    def __str__(self):
        result = f"{self.range}"
        if self.outcomes is not None and len(self.outcomes) > 0:
            result += f"{self.outcomes}"
        return result

    __repr__ = __str__


@dataclass
class Answer:
    outcomes: List[Outcome]
    constraint: Constraint
    cost: float = 0.0
    name: str = ""

    def __str__(self):
        if len(self.name) > 0:
            return self.name + f"{self.constraint}"
        else:
            output = f"{self.constraint}"
            if self.cost > 1e-7:
                output += f"(cost:{self.cost})"
            if len(self.outcomes) > 0:
                output += f"(outcomes:{self.outcomes})"
            return output

    __repr__ = __str__


@dataclass
class Query:
    answers: List[Answer]
    probs: List[float]
    cost: float = 0.0
    name: str = ""

    def __str__(self):
        if len(self.name) > 0:
            return self.name
        else:
            if self.cost < 1e-7:
                return f"answers: {self.answers}"
            else:
                return f"answers: {self.answers} (cost:{self.cost})"

    __repr__ = __str__


@dataclass
class QResponse:
    answer: Optional[Answer]
    indx: int
    cost: float


def possible_queries(
    ami: AgentMechanismInterface,
    strategy: "EStrategy",
    user: "User",
    outcome: "Outcome" = None,
) -> List[Tuple[Outcome, List["UtilityDistribution"], float]]:
    """Gets all queries that could be asked for that outcome until an exact value of ufun is found.

    For each ask,  the following tuple is returned:
    (outcome, query, cost)

    """
    user = copy.deepcopy(user)
    strategy = copy.deepcopy(strategy)

    def _possible_queries(outcome, strategy=strategy, ami=ami):
        queries_before = user.elicited_queries()
        utility_before = strategy.utility_estimate(outcome)
        lower_before, upper_before = _loc(utility_before), _upper(utility_before)
        n_before = len(queries_before)
        while True:
            u, _ = strategy.apply(user=user, outcome=outcome)
            if isinstance(u, float):
                break
        _qs = user.elicited_queries()[n_before:]

        # update costs
        s = 0.0
        qs = []
        for i, q in enumerate(_qs):
            qs.append((outcome, q.query, q.cost + s - user.cost))
            s += q.cost

        # # add possible other answers
        # for old_indx, _ in enumerate(qs):
        #     if strategy.strategy == 'exact':
        #         qs[old_indx] = (_[0], [_[1]], _[3], _[4], _[5])
        #         continue
        #     others = []
        #     if (_[1] - lower_before) > epsilon:
        #         others.append(UtilityDistribution(dtype='uniform', loc=lower_before, scale=_[1] - lower_before))
        #     others.append(UtilityDistribution(dtype='uniform', loc=_[1], scale=_[2]) if _[2] > 0 else _[1])
        #     end = (_[1] + _[2])
        #     if (upper_before - end) > epsilon:
        #         others.append(UtilityDistribution(dtype='uniform', loc=end, scale=upper_before - end))
        #     qs[old_indx] = (_[0], others, _[3], _[4], _[5])
        return qs

    if outcome is None:
        queries = []
        for outcome in ami.outcomes:
            queries += _possible_queries(outcome)
    else:
        queries = _possible_queries(outcome)
    return queries


def next_query(
    strategy: "EStrategy", user: "User", outcome: "Outcome" = None
) -> List[Tuple[Outcome, Query, float]]:
    """Gets the possible outcomes for the next ask with its cost.

    The following tuple is returned:
    (outcome, query, cost)

    """

    def _next_query(outcome, strategy=strategy):
        return outcome, strategy.next_query(outcome), user.cost_of_asking()

    if outcome is None:
        queries = []
        for outcome in strategy.outcomes:
            queries.append(_next_query(outcome))
    else:
        queries = [_next_query(outcome)]
    return queries


class CostEvaluator:
    def __init__(self, cost: float):
        self.cost = cost

    def __call__(self, query: Query, answer: Answer):
        return self.cost + query.cost + (answer.cost if answer.cost else 0.0)



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
Autogenerated by poetry-setup:
https://github.com/orsinium/poetry-setup
"""
# IMPORTANT: this file is autogenerated. Do not edit it manually.
# All changes will be lost after `poetry-setup` command execution.
# ----------------------------------------------------------------
# Always prefer setuptools over distutils
import re

# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    readme_txt = f.read()
with open(path.join(here, "HISTORY.rst"), encoding="utf-8") as f:
    history_txt = f.read()

long_description = "%s\n%s" % (
    re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub("", readme_txt),
    re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", history_txt),
)
# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
setup(
    # https://packaging.python.org/specifications/core-metadata/#name
    name="negmas",  # Required
    # https://www.python.org/dev/peps/pep-0440/
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.6.13",  # Required
    # https://packaging.python.org/specifications/core-metadata/#summary
    description="NEGotiations Managed by Agent Simulations",  # Required
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type="text/x-rst",  # Optional (see note above)
    url="https://github.com/yasserfarouk/negmas",  # Optional
    author="Yasser Mohammad",  # Optional
    author_email="yasserfarouk@gmail.com",  # Optional
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],  # Optional
    keywords=" ".join(
        ["negotiation", "mas", "multi-agent", "simulation", "AI"]
    ),  # Optional
    packages=find_packages(
        exclude=[
            "negmas/**/*.pyc",
            "nemgas/**/__pycache__",
            "config",
            "docs",
            "etc",
            "notebooks",
            "tests",
        ]
    ),  # Required
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "Click (>=6.0)",
        'dataclasses; python_version < "3.7"',
        "PyYAML",
        "progressbar2 (>=3.39)",
        "typing_extensions (>=3.7)",
        "pytest-runner (>=4.4)",
        "pandas (>=0.24.1)",
        "scipy (>=1.2)",
        "numpy (>=1.16)",
        "stringcase",
        "colorlog",
        "inflect",
        "matplotlib",
        "setuptools (>=40.8.0)",
        "tabulate",
        "typing",
        "tqdm (>=4.31.1)",
        "click-config-file",
        "dill",
        "seaborn",
        "sklearn",
        "networkx",
        "gif",
        "py4j",
    ],  # Optional
    extras_require={
        "elicitation": ["blist"],
        "visualization": ["flask", "dash", "dash-daq", "dash-bootstrap-components"],
        "gui": ["pyqt5"],
    },
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#dependencies-that-aren-t-in-pypi
    dependency_links=[],  # Optional
    # https://stackoverflow.com/a/16576850
    include_package_data=True,
    entry_points={  # Optional
        "console_scripts": ["negmas=negmas.scripts.app:cli", "negui=negmas.gui.app:cli"]
    },
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={"homepage": "https://github.com/yasserfarouk/negmas"},  # Optional
    python_requires=">=3.6",
)



import random

import pytest

from negmas import Issues, Issue, outcome_in_range, outcome_is_valid

from .fixtures import (
    issues,
    valid_outcome_dict,
    invalid_outcome_dict,
    valid_outcome_list,
    invalid_outcome_list,
    int_issues,
    cissue,
    dissue,
    sissue,
)


def test_dict_outcomes(issues, valid_outcome_dict, invalid_outcome_dict):
    assert outcome_is_valid(valid_outcome_dict, issues)
    assert not outcome_is_valid(invalid_outcome_dict, issues)


def test_list_outcomes(int_issues, valid_outcome_list, invalid_outcome_list):
    assert outcome_is_valid(valid_outcome_list, int_issues)
    assert not outcome_is_valid(invalid_outcome_list, int_issues)


def test_outcome_in_verious_ranges():
    or1 = {"price": (0.0, 2.0), "distance": [0.3, 0.4], "type": ["a", "b"], "area": 3}
    or2 = {"price": [(0.0, 1.0), (1.5, 2.0)], "area": [(3, 4), (7, 9)]}

    assert outcome_in_range({"date": "2018.10.4"}, or1)
    assert not outcome_in_range({"date": "2018.10.4"}, or1, strict=True)
    assert outcome_in_range({"area": 3}, or1)
    assert not outcome_in_range({"type": "c"}, or1)
    assert outcome_in_range({"type": "a"}, or1)
    assert not outcome_in_range({"date": "2018.10.4"}, or2, strict=True)
    assert not outcome_in_range({"area": 3}, or2)
    assert outcome_in_range({"area": 3.0001}, or2)
    assert not outcome_in_range({"area": 5}, or2)
    assert outcome_in_range({"price": 0.4}, or2)
    assert not outcome_in_range({"price": 1.2}, or2)
    assert not outcome_in_range({"price": 0.4, "area": 4}, or2)
    assert not outcome_in_range({"price": 0.4, "area": 10}, or2)
    assert not outcome_in_range({"price": 1.2, "area": 10}, or2)
    assert not outcome_in_range({"price": 1.2, "area": 4}, or2)
    assert outcome_in_range({"type": "a"}, or2)
    or1 = {"price": 10}
    assert outcome_in_range({"price": 10}, or1)
    assert not outcome_in_range({"price": 11}, or1)


def test_issues_construction():
    issues = Issues(price=(0.0, 1.0), cost=[1, 2, 3], delivery=["yes", "no"])
    assert len(issues.issues) == 3
    assert str(issues) == "price: (0.0, 1.0)\ncost: [1, 2, 3]\ndelivery: ['yes', 'no']"
    assert issues.is_infinite()
    assert not issues.is_finite()
    assert all(
        a == b for a, b in zip(issues.types, ["continuous", "discrete", "discrete"])
    )

    issues = Issues.from_single_issue(Issue(10, "issue"))
    assert len(issues.issues) == 1
    assert str(issues) == "issue: (0, 9)"

    issues = Issues(price=[2, 3], cost=[1, 2, 3], delivery=["yes", "no"])
    assert issues.is_finite()
    assert not issues.is_infinite()
    assert issues.cardinality == issues.num_outcomes == 2 * 3 * 2

    valid = issues.rand_valid()
    invalid = issues.rand_invalid()
    assert outcome_in_range(valid, issues.outcome_range)
    assert outcome_in_range(invalid, issues.outcome_range)


def test_outcome_constraint():
    r1 = {
        "price": [(0.0, 1.0), (3.0, 4.0)],
        "quantity": [1, 2, 4, 5],
        "delivery": "yes",
    }

    notr1 = {"price": [(3.01, 3.99), (4.01, 10.0)], "quantity": 3, "delivery": "no"}

    r2 = {"price": (3.01, 10.0), "quantity": [2, 3], "delivery": ["yes", "no"]}

    # r = OutcomeSpace(r1)


def test_from_outcomes():
    issues = Issues(price=[2, 3], cost=[1, 2, 3], delivery=["yes", "no"])
    found = Issue.from_outcomes(Issue.enumerate(issues.issues))
    for i, f in zip(issues.issues, found):
        assert i.name == f.name
        assert all(a == b for a, b in zip(sorted(i.values), f._values))

    issues = Issues(price=(1, 7), cost=(0, 5), delivery=["yes", "no"])
    found = Issue.from_outcomes(
        Issue.enumerate(issues.issues, max_n_outcomes=1000), numeric_as_ranges=True
    )
    for i, f in zip(issues.issues, found):
        v = sorted(i.values)
        assert i.name == f.name
        assert f._values[0] >= v[0] and f._values[1] <= v[1]


if __name__ == "__main__":
    pytest.main(args=[__file__])



"""
Common data-structures for supporting the Stacked Alternating Offers Protocol
"""
from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from negmas.common import AgentMechanismInterface, MechanismState
from negmas.outcomes import Outcome, ResponseType

__all__ = [
    "SAOResponse",
    "SAOState",
    "SAOAMI",
]


@dataclass
class SAOResponse:
    """A response to an offer given by an agent in the alternating offers protocol"""

    response: ResponseType = ResponseType.NO_RESPONSE
    outcome: Optional["Outcome"] = None


@dataclass
class SAOState(MechanismState):
    current_offer: Optional["Outcome"] = None
    current_proposer: Optional[str] = None
    current_proposer_agent: Optional[str] = None
    n_acceptances: int = 0
    new_offers: List[Tuple[str, "Outcome"]] = field(default_factory=list)
    new_offerer_agents: List[str] = field(default_factory=list)


@dataclass
class SAOAMI(AgentMechanismInterface):
    end_on_no_response: bool = True
    publish_proposer: bool = True
    publish_n_acceptances: bool = False
